"""

    LDAPUtilities.py
    $Author: payne $
    $Date: 2003/08/16 01:21:51 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    A repository for useful utilities that are not Zope-aware.

"""

__version__ = '$Revision: 1.6 $'[11:-2]

TRUE = 1
FALSE = 0

from copy import deepcopy
from LDAPExceptions import INVALID_DN_SYNTAX
from string import join, lower, split, strip

try:
    import _ldap
    if hasattr( _ldap, 'initialize' ):
        raise ImportError
except ImportError:
    import ldap
    _ldap = ldap

# Stolen from Jens' LDAPUserFolder:
#
PY_LDAP_VERSION = getattr( _ldap, '__version__', '1.10alpha3' )

def explode( distinguishedName, notypes=0 ):
    """
    Wrap the _ldap.explode_dn function
    """
    return _ldap.explode_dn( distinguishedName, notypes )

def wildDN( dn ):
    """
    When a dn is to be used as the value of an attribute search,
    use this method to insert wildcards ('*') wherever white space
    in the dn should not make a difference.
    """
    # Example:
    #     'ou=ST ScI,o=test'  => 'ou*=*ST ScI*,*o*=*test'

    rdns = map( lambda x: join( split(x,'='), '*=*' ), explode(dn) )
    return join( rdns, '*,*' )

def DNNormalize( distinguishedName ):
    """
        Normalize a distinguished name into a form suitable for
        comparison with another normalized distinguished name.
        Lowercase and remove all extraneous spaces.
    """
    return lower( join( explode(distinguishedName), ',') )

def DNsAreEquivalent( DN1, DN2 ):
    """
        Returns true if distinguished names match.  Ignore case and
        white space around keys and values.
    """
    if str(DNNormalize( DN1 )) == str( DNNormalize( DN2 )):
        return TRUE
    else:
        return FALSE

def getNameFromDN( distinguishedName ):
    """
        Returns the RDN from a distinguished name with the
        attribute type removed.

        Example

            getNameFromDN( 'attributeName=attributeValue,ou=orgUnit,o=organization')

        returns the string 'attributeValue'
    """
    assert type(distinguishedName) == type('')
    try:
        rdn = explode( distinguishedName, 1 )[0]
        return rdn
    except:
        raise INVALID_DN_SYNTAX
    
def getParentFromDN( distinguishedName ):
    """
    Return the "parent" of the given name by taking off the RDN
    """
    assert type(distinguishedName) == type('')
    try:
        parent = join( explode( distinguishedName)[1:], ',' )
        return parent
    except:
        raise INVALID_DN_SYNTAX
    
def listsAreEquivalent( list1, list2 ):
    """
        Returns True if all elements in the first list are contained
        in the second and vice versa.
    """
    assert (type(list1) == type([])) and (type(list2) == type([]))

    tmp1 = listCopy(list1)
    tmp2 = listCopy(list2)

    for item in list1:
        tmp1.remove(item)
        if item in tmp2:
            tmp2.remove(item)
        else:
            return 0

    if len(tmp1) > 0:
        return 0

    if len(tmp2) > 0:
        return 0
    else:
        return 1

def listCopy( list ):
    """
        Returns a copy of the input list.
    """
    assert type(list) == type([])
    return deepcopy( list )

def list1ContainsList2( list1, list2 ):
    """
        Return TRUE if all the items in list2 are in list1.
    """
    for item in list2:
        if not item in list1:
            return FALSE
    return TRUE

def mergeLists( list1, list2 ):
    """
        Merges two lists without duplicating
        values that appear in both.
    """
    assert type(list1) == type([])
    assert type(list2) == type([])
    
    mergedList = list1 + filter( lambda i,l=list1: i not in l, list2 )
    return mergedList
