"""
    LDIFObjectParser.py
    $Author: payne $
    $Date: 2003/08/16 01:21:51 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    The LDIFObjectParser provides methods for parsing an LDIF
    object definition, and for converting the result back to LDIF.

    The general form of an LDIF object definition is as follows:
 
    ( oid NAME 'objectClassName' DESC 'object class description'
      SUP superclassName OBJECTTYPE MUST must-contain-list
      MAY may-contain-list X-ORIGIN 'who defined this' )
 
    Each field is as follows:
 
        oid  - required.  A string of the form n.m.x.y.z...
        NAME - required.  Always surrounded with single quotes.
        DESC - optional.  If it is given, the value is surrounded
                          by single quotes.
        SUP  - optional.  Required for all but the class 'top'.
                          Defines the class that this class
                          subclasses.  The superclass name is not
                          enclosed in quotes.
        OBJECTTYPE - required.  Is one of the following strings:
                          ABSTRACT, AUXILIARY, STRUCTURAL
        MUST - optional   A list of attributes that must be given
                          values any time an object of this class
                          is created.  The list is either a singleton,
                          e.g. cn or is a list separated by dollar
                          signs and enclosed in parentheses.
        MAY  - optional   A list of attributes that may be given
                          values any time an object of this class
                          is created.  The list is either a singleton,
                          e.g. cn or is a list separated by dollar
                          signs and enclosed in parentheses.
        X-ORIGIN - optional.  If given, the value is surreounded
                          iin single quotes.

"""

__version__ = '$Revision: 1.10 $'[11:-1]

from string import find, join, split

class LDIFObjectParser:

    def __init__( self, LDIFExpression = None):

        if LDIFExpression:

            self._parseLDIF( LDIFExpression )

    def getLDIF(self):
        """
            Returns the LDIF string defining the objectClass.
        """
        req = self.oid + " NAME '" + self.name + "' "

        if self.description:
            desc = "DESC '" + self.description + "' "
        else:
            desc = ""

        if self.superClass:
            super = "SUP " + self.superClass + " "
        else:
            super = ""

        if self.mustContain:
            if len(self.mustContain) == 1:
                must = "MUST " + self.mustContain[0] + " "
            else:
                must = "MUST ( " + join( self.mustContain, " $ ") + " ) "
        else:
            must = ""

        if self.mayContain:
            if len(self.mayContain) == 1:
                may = "MAY " + self.mayContain[0] + " "
            else:
                may = "MAY ( " + join( self.mayContain, " $ ") + " ) "
        else:
            may = ""

        if self.origin:
            origin = "X-ORIGIN '" + self.origin + "'"

        return "( " + req + desc + super + self.objectType + " " + must + may + origin + " )"
        
        
    def _parseLDIF( self, LDIFExpression ):
        
        # Strip surrounding parentheses if they are there.

        if LDIFExpression[0:2] == '( ':
            ldif = LDIFExpression[2:-1]
        else:
            ldif = LDIFExpression

        i = find(ldif,' ')
        self.oid = ldif[0:i]


        i = find(ldif,"NAME")
        if i == -1:
	    print "need NAME in LDIF\n"
	    print "%s\n" % ldif
	    assert 0

        i = find(ldif,"'", i)
        j = find(ldif,"'", i+1)
        self.name = ldif[i+1: j]

        i = find(ldif,"DESC")
        if i != -1:
            i = find(ldif,"'", i)
            j = find(ldif,"'", i + 1)
            self.description = ldif[i+1:j]
        else:
            self.description = ""

        i = find(ldif,"SUP")
        if i != -1:
            i = find(ldif," ", i)
            j = find(ldif," ", i + 1)
            self.superClass = ldif[i+1:j]
        else:
            self.superClass = ""

        if find(ldif,"STRUCTURAL" ) != -1:
            self.objectType = "STRUCTURAL"
        elif find(ldif,"ABSTRACT") != -1:
            self.objectType = "ABSTRACT"
        elif find(ldif,"AUXILIARY") != -1:
            self.objectType = "AUXILIARY"
        else:
	    print "need STRUCTURAL, ABSTRACT or AUXILIARY in LDIF\n"
	    print "%s\n" % ldif
            assert 0

        self.mustContain = self._getAttrList( ldif, "MUST" )
        self.mayContain = self._getAttrList( ldif, "MAY" )

        i = find(ldif,"X-ORIGIN" )
        if i != -1:
            i = find(ldif, "'", i)
            j = find(ldif, "'", i+1)
            self.origin = ldif[i+1 : j]
        else:
            self.origin = ''

    def _getAttrList( self, ldif, attrType ):
        
        i = find(ldif,attrType)
        if i == -1:
            return []
        i = find(ldif," ", i+1)
        if ldif[i+1:i+2] == "(":
            i = i + 2
            j = find(ldif, " )", i)
        else:
            j = find(ldif, " ", i+1 )
        tmp = ldif[i+1 : j]
        return split(tmp," $ ")
