"""

    ManagedRoles.py
    $Author: payne $
    $Date: 2003/08/16 01:21:51 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    The role management classes.  A single, static, persistent
    ManagedRoles object is instantiated with default values when
    a new LDAP manager object is created.  Dynamic, volatile
    ZRoles objects are created whenever roles need to be changed.

    There are two kinds of roles.  There are five permanent roles
    provided for five special functionalities.  These five roles
    have the canonical names 'SELF', 'ANON', 'AUTH', 'MNGR', and
    'IFALLOWED'.  The _permanentRoles attribute is a dictionary
    of dictionaries.  The keys to the outer dictionary are the
    canonical names.  The keys to each of the inner dictionaries
    are 

        roleName        a descriptive name, and
        description     a short descriptive text.

    The role names may be changed and reordered, but their
    descriptions and functionality cannot.  The descriptive
    names and descriptions are used on the forms for managing
    roles, but the canonical names are used internally.  In
    particular, the Readers and Writers for attributes are
    stored as lists of canonical names.

    The other roles are configured roles.  These are roles that
    match groups in the LDAP directory.  The groups must exist,
    and be findable using the groupSearchBase and groupSearchScope
    specified under the configuration tab.
"""

__version__ = '$Revision: 1.13 $'[11:-2]

from copy import deepcopy
from LDAPUtilities import _ldap, DNNormalize, DNsAreEquivalent, \
    getNameFromDN, listsAreEquivalent, listCopy, list1ContainsList2, \
    mergeLists, TRUE, FALSE
from SchemaDefaults import *

class ManagedRoles:

    def __init__(self):

        self._canonicalRoleNames = listCopy( defaultCanonicalRoleNames )
        self._permanentRoles = deepcopy(defaultPermanentRoles)
        self._configuredRoleNames = listCopy( defaultConfiguredRoleNames )

    def addRole( self, role ):
        """
            Add role to the list of configured role names.
        """
        if not role in self._configuredRoleNames:
            self._configuredRoleNames.append( role )

    def delRole( self, role ):
        """
            Remove role from the list of configured role names.
        """
        assert role in self._configuredRoleNames
        self._configuredRoleNames.remove( role )

    def getCanonicalRoleNames( self ):
        """
            Returns the (ordered) list of canonical role names.
        """
        return listCopy( self._canonicalRoleNames + \
                         self._configuredRoleNames )

    def getConfiguredRoleNames( self ):
        """
            Returns the list of configured role names.
        """
        return listCopy( self._configuredRoleNames )

    def getPermanentRoles( self ):
        """
            Returns the (ordered) list of permanent role names.
        """
        return listCopy( self._canonicalRoleNames )

    def getPermanentRoleDescriptions( self ):
        """
            Returns the ordered list of permanent role names.
        """
        return map( lambda i,s=self: s.getRoleDescriptionFromCanonicalName(i), 
                    self._canonicalRoleNames )

    def getRoleDescriptionFromCanonicalName( self, canonicalName ):
        """
            Returns the role description for the given canonicalName
        """
        return self._permanentRoles[canonicalName]['description']

    def getRoleNameFromCanonicalName( self, canonicalName ):
        """
            Returns the descriptive name for the given canonicalName
        """
        if canonicalName in self._permanentRoles.keys():
            return self._permanentRoles[canonicalName]['roleName']
        elif canonicalName in self.getConfiguredRoleNames(): 
            return canonicalName
        else:
            return ""

    def getPermanentRoleNames( self ):
        """
            Returns the ordered list of permanent role names.
        """
        return map( lambda i,s=self: s.getRoleNameFromCanonicalName(i),
                    self._canonicalRoleNames )

    def getRoleNames( self ):
        """
            Returns the list of role names: the five fixed roles
            followed by any configured roles.
        """
        return listCopy( self.getPermanentRoleNames() +
                         self.getConfiguredRoleNames() )

    def setCanonicalRoleOrder( self, roles ):
        """
            Reorder the canonical role names.
        """
        assert listsAreEquivalent( roles, self._canonicalRoleNames )
        self._canonicalRoleNames = roles

    def setPermanentRoleNames( self, roles ):
        """
            Set permanent role names.  No re-ordering allowed!
        """
        assert type( roles ) == type([])
        assert len( roles ) == len( self._canonicalRoleNames )

        for i in range( 0, len( self._canonicalRoleNames ) ):
            self._permanentRoles[self._canonicalRoleNames[i]]['roleName'] = \
                roles[i]

