"""

    ManagedSchema.py
    $Author: payne $
    $Date: 2003/12/08 02:58:14 $

    This class encapsulates the LDAP schema for a particular objectCategory
    ('person', 'group', 'orgUnit').  A singleton for each objectCategory is
    created in the LDAPDirectory __init__ method, invoked from the
    STLDAPManager __init__ method whenever a new directory manager is
    instantiated.

    A ManagedSchema contains a dictionary of attributes, where the key
    is the ldapAttributeName of the attribute, and the value is a 
    ManagedAttributeSchema object, which encapsulates all of the 
    information about that attribute.

    It also contains:

      o A list of "Controllers," i.e., roles allowed to add and remove
        objects of this type from the directory,

      o A list of "Display Categories," which is a list of category names
        used to organize the attributes.  Each ManagedAttributeSchema
        object contains a list of display categories to which that
        attribute belongs,

      o The naming attribute for this category of object, e.g., "ou"
        for object category "orgUnit" and "cn" for object category
        "group,"

      o A searchFilter that limits search results to this type of
        object, e.g., "objectclass=person", and can be configured to
        further restrict results (for example, we add a test that
        records are active),

      o A list of searchKeys, ldapAttributeNames of the attributes that
        will be searched, and a matching searchKeyNames list of friendly
        names (name order matters, determining the layout of search
        result display tables), and

      o A searchSortKey, the ldapAttributeName of the attribute used
        for sorting search results.

      o A browseFormat string, used to format the browseKeys values
        into a label in the browse view of the directory.

      o A list of browseKeys, ldapAttributeNames of the attributes that
        will be used to label objects on browse view trees.

      o A browseSortKey, the ldapAttributeName of the attribute used
        to sort entries at each node of the browse view of the 
        directory.

    A ManagedSchema is instantiated with defaults taken from the
    SchemaDefaults module.  In the Zope product, a ZSchema object makes
    a connection to the LDAP directory to update/obtain all of the 
    information carried by this object.

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

"""

__version__ = '$Revision: 1.39 $'[11:-2]

from copy import deepcopy
from LDAPExceptions import *
from LDAPUtilities import _ldap, DNNormalize, DNsAreEquivalent, \
    getNameFromDN, listsAreEquivalent, listCopy, list1ContainsList2, \
    mergeLists, TRUE, FALSE
from ManagedAttributeSchema import ManagedAttributeSchema
from SchemaDefaults import *

class ManagedSchema:

    def __init__(self,
                 defaultDictionary,
                 defaultDisplayCategories,
                 defaultNamingAttribute,
                 defaultControllers,
                 defaultSearchFilter,
                 defaultSearchKeys,
                 defaultSearchSortKey,
                 browseFormat=None,
                 browseKeys=None,
                 browseSortKey=None,
                 personsOUAttr=None,
		 permissionName=''):

        self._pointsToTargets = defaultPointsToTargets
        
        self._DisplayCategories = deepcopy(defaultDisplayCategories)
        self._Controllers = deepcopy(defaultControllers)

        self._attributeCollection = {}
        self._allowedLDAPAttrNames = defaultDictionary.keys()
        self._namingAttribute = defaultNamingAttribute
        if personsOUAttr is not None:
            self._personsOUAttr = personsOUAttr
        
        for ldapAttributeName in self._allowedLDAPAttrNames:
            attributeSchema = ManagedAttributeSchema(ldapAttributeName,
                                          defaultDictionary[ldapAttributeName])
            attributeSchema.setManaged(TRUE)
            self._attributeCollection[ldapAttributeName] = attributeSchema

        #   Now, because OpenLDAP allows for attributes to be subclassed 
        #   from other attributes, the attribute collection may contain 
        #   ManagedAttributeSchema objects that have a specified superior 
        #   but no other attributes.  Let's fix that now by finding 
        #   attributes with a superior, finding the superior attribute
        #   schema object, and passing it back to the target attribute's
        #   inheritsFrom method.

        for ldapAttributeName in self._allowedLDAPAttrNames:
            attributeSchema = self._attributeCollection[ldapAttributeName]
            if attributeSchema.getSup():
                superiorAttrName = attributeSchema.getSup()
                while superiorAttrName:
                    superiorAttr = self._attributeCollection[superiorAttrName]
                    attributeSchema.inheritsFrom( superiorAttr )
                    superiorAttrName = superiorAttr.getSup()

        #   Filter out search keys not in the allowedLDAPAttrNames list,
        #   and then build the list of searchKeyNames.

        self._searchFilter = defaultSearchFilter
        self._searchSortKey = defaultSearchSortKey
        self._searchKeys = deepcopy( defaultSearchKeys )
        self.setSearchKeyFriendlyNames()

	#   The controllers need write access to the namingAttribute
	#   for entry renaming to work under OpenLDAP.

	namingAttributeSchema = self._attributeCollection[self._namingAttribute]
	namingAttributeSchema.addWriters( self._Controllers )

        #   Save the browse view parameters.

        self._browseFormat = browseFormat
        self._browseKeys = deepcopy(browseKeys)
        self._browseSortKey = browseSortKey

	#   The Zope permission that controls managing this type of
	#   object.  Now, it gets set later.

	self._permissionName = permissionName

    def addBrowseKeys( self, keys ):
        """
        Add keys to the list of browse keys.
        """
        for key in keys:
            if key not in self.getBrowseKeys():
                self._browseKeys = self.getBrowseKeys() + [ key ]

    def addController(self, role):
        """
        Adds a role to the list of roles that are allowed to
        create and delete LDAP objects of this schema.
        """
        if not role in self._Controllers:
            self._Controllers.append(role)

    def addDisplayCategory( self, category ):
        """
        Add category to the list of display categories.
        """
        if not category in self._DisplayCategories:
            self._DisplayCategories.append( category )

    def addSearchKeys( self, keys ):
        """
        Add keys to the list of search keys.
        """
        for key in keys:
            if key not in self._searchKeys:
                self._searchKeys = self._searchKeys + [ key ]
                friendlyName = self.getAttributeSchema( key ).getFriendlyName()
                self._searchKeyNames[key] = friendlyName

    def delController(self, role):
        """
        Deletes a role from the list of roles that are allowed to
        create and delete LDAP objects of this schema.
        """
        if role in self._Controllers:
            self._Controllers.remove(role)

    def delDisplayCategory( self, category ):
        """
        Remove category from the list of display categories.
        """
        assert category in self._DisplayCategories
        self._DisplayCategories.remove( category )

    def getAllowedAttributeNames(self):
        """
        Returns a list of the names of all allowed attributes.
        """
        return deepcopy( self._allowedLDAPAttrNames )

    def getAttributeCollection(self):
        """
        Returns a copy of the entire dictionary of
        ManagedAttributeSchema objects.
        """
        return self._attributeCollection.copy()

    def getAttributeSchema(self, ldapName):
        """
        Returns a named ManagedAttributeSchema object.
        """
        assert ldapName in self._allowedLDAPAttrNames
        return self._attributeCollection[ldapName]

    def getAvailableAttributeNames(self):
        """
        Returns a list of the names of all allowed attributes which
        are not managed.
        """
        available = filter( lambda i,ac=self._attributeCollection: 
                            not ac[i].getManaged(), 
                            self._allowedLDAPAttrNames)
        available.sort()
        return available

    def getBrowseFormat(self):
        """
        Returns the browse format string for this object category.
        """
        if hasattr(self,"_browseFormat"):
            return self._browseFormat
        else:
            return '%s'

    def getBrowseKeys(self):
        """
        Returns the list of browse key ldapAttributeNames.
        """
        if hasattr(self,"_browseKeys") and self._browseKeys:
            return self._browseKeys 
        else:
            return []

    def getBrowseSortKey(self):
        """
        Returns the ldapAttributeName of the attribute on which to
        sort browse results.
        """
        if hasattr(self,'_browseSortKey'):
            return self._browseSortKey
        else:
            return None

    def getControllers(self):
        """
        Returns a list of roles that are allowed to create and
        delete LDAP objects that follow this schema.
        """
        return self._Controllers

    def getDisplayCategories(self):
        """
        Returns the list of display categories.
        """        
        return deepcopy( self._DisplayCategories )

    def getManagedAttributeNames(self):
        """
        Returns a list of the names of all managed attributes.
        """
        managed = filter( lambda i,ac=self._attributeCollection:
                          ac[i].getManaged(),
                          self._allowedLDAPAttrNames )
        return managed

    def getManagedAttributeNamesByCategory(self, displayCategory):
        """
        Returns a list of managed attribute names in the given
        display category and for which the user has at least
        read access.
        """
        managed = \
            filter( lambda i,ac=self._attributeCollection,dc=displayCategory:
                    dc in ac[i].getDisplayCategories(),
                    self.getManagedAttributeNames() )
        return managed

    def getManagedNotRequiredAttributeNames(self):
        """
        Returns a list of the names of all managed attributes which
        are not required.
        """
        managedNR = filter( lambda i,ac=self._attributeCollection:
                            not ac[i].getRequired(),
                            self.getManagedAttributeNames() )
        managedNR.sort()
        return managedNR

    def getNamingAttribute(self):
        """
        Returns the naming attribute for this object schema.
        """
        return self._namingAttribute

    def getPermissionName(self):
	"""
	Return the name of the Zope permission that controls managing
	this kind of object.
	"""
	return self._permissionName

    def getPersonsOUAttr(self):
        """
        Return the name of the attribute to examine to find a person's
        organizational unit.
        """
        return self._personsOUAttr

    def getPointsToTargets(self):
        """
        Returns a list of pointsTo target values.
        """
        return self._pointsToTargets

    def getReadableAttributesByRole( self, role ):
        """
        Returns a list of managed attributes readable, but not
        writeable, by the given role.
        """
        writeable = self.getWriteableAttributesByRole( role )
        readable = filter( lambda i,r=role,s=self,w=writeable: 
                           (r in s.getAttributeSchema(i).getReaders()) 
                           and (not i in w), 
                           self.getManagedAttributeNames() )
        readable.sort()
        return readable

    def getRequiredAttributeNames(self):
        """
        Returns a list of the names of all required attributes for this
        object.
        """
        required = filter( lambda i,ac=self._attributeCollection:
                           ac[i].getRequired(),
                           self.getManagedAttributeNames() )
        required.sort()
        return required

    def getSearchFilter(self):
        """
        Returns the search filter for this object category.
        """
        return self._searchFilter

    def getSearchKeyName( self, key ):
        """
        Returns the search key name for the specified search key
        ldapAttributeName.
        """
        return self._searchKeyNames[key]

    def getSearchKeyNames(self, keys ):
        """
        Returns the list of search key names in the order of the
        specified list of search key ldapAttributeNames.  The
        keys are passed in because some of the columns may not
        be readable by this user.
        """
        searchKeyNames = map( lambda i,s=self._searchKeyNames: s[i], keys )
        return searchKeyNames

    def getSearchKeys(self):
        """
        Returns the list of search key ldapAttributeNames.
        """
        return self._searchKeys 

    def getSearchSortKey(self):
        """
        Returns the ldapAttributeName of the attribute on which to
        sort search results.
        """
        return self._searchSortKey

    def getWriteableAttributesByRole( self, role ):
        """
        Returns a list of managed attributes writeable by the given role.
        """
        writeable = filter( lambda i,r=role,s=self: 
                            r in s.getAttributeSchema(i).getWriters(),
                            self.getManagedAttributeNames() )
        writeable.sort()
        return writeable

    def manageAttributeSchema( self, ldapName ):
        """
        Add the named ManagedAttributeSchema object to the
        list of managed attributes.
        """
        assert type(ldapName) == type('')
        if not ldapName in self._allowedLDAPAttrNames:
            raise NO_SUCH_ATTRIBUTE
        self._attributeCollection[ldapName].setManaged( TRUE )

    def setAllowedAttributeNames( self, allowedNames ):
        """
        Set the (sorted) list of allowed attribute names.
        """
        allowedNames.sort()
        self._allowedLDAPAttrNames = allowedNames

    def setBrowseFormat( self, browseFormat ):
        """
        Set the browse format.
        """
        self._browseFormat = browseFormat

    def setBrowseKeys( self, browseKeys ):
        """
        Set the list of browse key ldapAttributeNames.
        """
        self._browseKeys = browseKeys

    def setBrowseSortKey( self, sortKey ):
        """
        Set the ldapAttributeName of the attribute on which to
        sort browse results.
        """
        self._browseSortKey = sortKey

    def setDisplayCategories( self, categories ):
        """
        Set display categories.
        """
        if type( categories ) == type(''):
            categories = [categories,]
        assert type( categories ) == type([])
        self._DisplayCategories[0:len(self._DisplayCategories)] = categories

    def setNamingAttribute( self, attributeName ):
        """
        Set the naming attribute for this object schema.
        """
        assert type(attributeName) == type('')
        if not attributeName in self._allowedLDAPAttrNames:
            raise NO_SUCH_ATTRIBUTE
        self._namingAttribute = attributeName

    def setPermissionName( self, permissionName ):
	"""
	Set the name of the Zope permission that controls managing
	this kind of object.
	"""
	self._permissionName = permissionName

    def setPersonsOUAttr( self, attributeName ):
        """
        Set the person's OU indicating attribute for this object schema.
        """
        assert type(attributeName) == type('')
        if not (attributeName == 'dn' or \
         attributeName in self._allowedLDAPAttrNames):
            raise NO_SUCH_ATTRIBUTE
        self._personsOUAttr = attributeName

    def setSearchFilter( self, searchFilter ):
        """
        Set the search filter.
        """
        self._searchFilter = searchFilter

    def setSearchKeyFriendlyNames(self):
        """
        Set up the search key friendly names dictionary by looking 
        up friendly names for search key ldapAttributeNames.
        """
        self._searchKeyNames = {}
        for key in self.getSearchKeys():
            if key in self._allowedLDAPAttrNames:
                friendlyName = self.getAttributeSchema( key ).getFriendlyName()
                self._searchKeyNames[key] = friendlyName

    def setSearchKeyName( self, key, keyName ):
        """
        Set the search key name for the specified key to be keyName.
        """
        self._searchKeyNames[key] = keyName

    def setSearchKeys( self, searchKeys ):
        """
        Set the list of search key ldapAttributeNames.
        """
        self._searchKeys = searchKeys

    def setSearchSortKey( self, sortKey ):
        """
        Set the ldapAttributeName of the attribute on which to
        sort search results.
        """
        self._searchSortKey = sortKey

    def unmanageAttributeSchema( self, ldapName ):
        """
        Remove the named ManagedAttributeSchema object from the
        list of managed attributes.
        """
        assert self._attributeCollection.has_key(ldapName)

        if self._attributeCollection[ldapName].getRequired():
            raise REQUIRED_ATTRIBUTE

        self._attributeCollection[ldapName].setManaged( FALSE )
