"""

    STLDAPManager.py
    $Author: payne $
    $Date: 2004/05/20 19:12:19 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

ST LDAP Manager class.

A class that lets users authenticate against an LDAP directory, and to
search, browse, and modify the directory to an extent that depends on
their roles.

"""
__version__ = '$Revision: 1.55 $'[11:-2]

from Connection import AnonymousUser, AuthenticatedUser
from LDAPDirectory import LDAPDirectory
from LDAPExceptions import *
from LDAPUtilities import _ldap, TRUE
from SchemaDefaults import *
from ZLDAPTree import ZLDAPTree
from ZObject import ZObject
from ZReports import ZReports
from ZRoles import ZRoles
from ZSchema import ZSchema

from AccessControl import ClassSecurityInfo, getSecurityManager
from AccessControl.Role import RoleManager
from Globals import DTMLFile, InitializeClass
from OFS.SimpleItem import SimpleItem
from zExceptions import Unauthorized

class STLDAPManager( ZLDAPTree, RoleManager, SimpleItem ):
    """
    A persistent object for managing objects in an LDAP directory.  
    Anonymous users may search, and authenticated users may update 
    the directory, limited by their roles.
    """

    meta_type = "LDAP Directory Manager"

    manageGroupPermission =   'LDAP Manage Groups'
    manageOrgUnitPermission = 'LDAP Manage Org Units'
    managePeoplePermission =  'LDAP Manage People'
    manageSchemaPermission =  'LDAP Manage Schema'

    manage_options = (
        ( 
          { 'label'  : 'People',
            'action' : 'manage_people',
            'help'   : ('STLDAPManager', 'People.stx' ) },
          { 'label'  : 'Org Units',
            'action' : 'manage_orgUnits',
            'help'   : ('STLDAPManager', 'OrgUnits.stx' ) },
          { 'label'  : 'Groups',
            'action' : 'manage_groups',
            'help'   : ('STLDAPManager', 'Groups.stx' ) },
          { 'label'  : 'Configuration', 
            'action' : 'manage_configure',
            'help'   : ('STLDAPManager', 'Add.stx' ) },
          { 'label'  : 'Roles',
            'action' : 'manage_roles',
            'help'   : ('STLDAPManager', 'Roles.stx' ) },
          { 'label'  : 'Schema',
            'action' : 'manage_schema',
            'help'   : ('STLDAPManager', 'Schema.stx' ) },
          { 'label'  : 'Searching & Browsing',
            'action' : 'manage_searching',
            'help'   : ('STLDAPManager', 'Searching.stx' ) },
          { 'label'  : 'Reports',
            'action' : 'manage_reports',
            'help'   : ('STLDAPManager', 'Reports.stx' ) }, 
          ) + \
	    SimpleItem.manage_options[:1]
        )

    security = ClassSecurityInfo()

    security.declareProtected( 'View', 'index_html')
    security.declareProtected( 'View', 'getUser')
    security.declareProtected( 'View', 'login')
    security.declareProtected( 'View', 'logout')
    security.declareProtected( 'View', 'showSearchResults')
    security.declareProtected( 'View', 'search')
    security.declareProtected( 'View', 'orgChart')
    security.declareProtected( 'View', 'viewJpegPhoto')
    security.declareProtected( manageGroupPermission, 'manage_groups' )
    security.declareProtected( manageOrgUnitPermission, 'manage_orgUnits')
    security.declareProtected( managePeoplePermission, 'manage_people' )
    security.declareProtected( manageSchemaPermission, 'manage_reports')
    security.declareProtected( manageSchemaPermission, 'manage_roles')
    security.declareProtected( manageSchemaPermission, 'manage_schema')
    security.declareProtected( manageSchemaPermission, 'manage_searching')
    security.declareProtected( manageSchemaPermission, 'manage_configureForm')
    security.declareProtected( manageSchemaPermission, 'manage_configure')
    security.declareProtected( manageSchemaPermission, 'manage_access')
    security.declareProtected( manageSchemaPermission, 'manage_undo_transactions')
    security.declareProtected( manageSchemaPermission, 'undoable_transactions')
    security.declareProtected( manageSchemaPermission, 'manage_UndoForm')

    manage_configureForm = DTMLFile('dtml/configure', globals())
    manage_main = manage_configureForm
    orgChartForm = DTMLFile('dtml/orgChart', globals())
    searchForm = DTMLFile('dtml/search', globals())
    viewForm = DTMLFile('dtml/view', globals())
    manage_UndoForm=DTMLFile('dtml/undo', globals(), PrincipiaUndoBatchSize=20,
        first_transaction=0, last_transaction=20 )

    def __init__( self, id, title, ldapServer, directoryBase, personSearchBase, 
                  orgUnitBrowseBase, groupBrowseBase, zopeRoleSearchBase, 
                  zopeRoleSearchScope, personNamingAttribute, 
                  personsOUAttr, givenAttr, groupObjectClass,
                  REQUEST=None ):

        self.id = id
        self.title = title
        self.LDAPDirectory = \
            LDAPDirectory( ldapServer, directoryBase, personSearchBase, 
                           orgUnitBrowseBase, groupBrowseBase, 
                           zopeRoleSearchBase, zopeRoleSearchScope,
                           personNamingAttribute, personsOUAttr,
                           givenAttr, groupObjectClass )
        user = self.getUser()
        self.schema = self.LDAPDirectory._personManagedSchema

        #   For each object category, create a temporary ZSchema object, 
        #   and use it to update the schema with information from LDAP,
        #   and to sync the list of object controllers with the permission
        #   associated with that category.

        zSchema = ZSchema( user, {'objectCategory':'person'} )
        zSchema = zSchema.__of__( self )
        zSchema.updateSchemaFromLDAP(user)
        zSchema.setPermissionName( self.managePeoplePermission )
        zSchema._assignRolesToPermission(roles=defaultManagePeopleRoles)

        zSchema = ZSchema( user, {'objectCategory':'orgUnit'} )
        zSchema = zSchema.__of__( self )
        zSchema.updateSchemaFromLDAP(user)
        zSchema.setPermissionName( self.manageOrgUnitPermission )
        zSchema._assignRolesToPermission(roles=defaultManageOrgUnitRoles)

        zSchema = ZSchema( user, {'objectCategory':'group'} )
        zSchema = zSchema.__of__( self )
        zSchema.updateSchemaFromLDAP(user)
        zSchema.setPermissionName( self.manageGroupPermission )
        zSchema._assignRolesToPermission(roles=defaultManageGroupRoles)

	for role in defaultManageSchemaRoles:
	    self._addRole( role )
	self.manage_permission( self.manageSchemaPermission,
				defaultManageSchemaRoles )

    def index_html( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
	an anonymous user or a user that logged in but doesn't have a
	role that allows seeing the People tab.
        """
        user = self.getUser()
	if isinstance( user, AuthenticatedUser ) and 'AUTH' in user.getRoles():
            REQUEST.set( 'objectCategory', 'person' )
            zObject = ZObject( user, REQUEST, message )
            zObject = zObject.__of__( self )
            return zObject.showObjectForm()
	elif REQUEST.has_key('objectFormID') or \
	 REQUEST.has_key('distinguishedName'):
	    # A trick for browsers that don't send authentication
	    # unless asked for it (e.g., Mozilla).
	    return self.login(REQUEST,REQUEST.RESPONSE)
        else:
            REQUEST.set( 'user', user )
            if not REQUEST.has_key('formID'):
                REQUEST.set( 'formID', 'search' )
            return self.viewForm( self, REQUEST )

    def login(self, REQUEST, RESPONSE):
        """ Allow user to authenticate. """
        user = self.getUser()
	if isinstance( user, AuthenticatedUser ):

	    # Redirect a user to the leftmost accessible tab

	    peopleRoles = map( lambda i: i['name'],
		filter( lambda i: i['selected']=='SELECTED',
		self.rolesOfPermission(self.managePeoplePermission)))
	    for role in user.getRoles():
		if role in peopleRoles:
            	    url = REQUEST.get( 'URL1' ) + '/manage_people'
            	    RESPONSE.redirect( url )
		    return

	    # If none is available, go back to the index_htmle page,
	    # with a QUERY_STRING, used to trick browsers that don't
	    # send authentication unless they have to.

            url = REQUEST.get( 'URL1' ) + '/index_html?objectFormID=search'
            RESPONSE.redirect( url )
	    return

	elif isinstance( user, AnonymousUser ):

	    # Do inherited Zope roles do any good?

	    schemaRoles = map( lambda i: i['name'],
		filter( lambda i: i['selected']=='SELECTED',
		self.rolesOfPermission(self.manageSchemaPermission)))
	    for role in user.getRoles():
		if role in schemaRoles:
            	    url = REQUEST.get( 'URL1' ) + '/manage_configure'
            	    RESPONSE.redirect( url )
		    return

	# Otherwise, raise Unauthorized and make user authenticate.

        realm = RESPONSE.realm
	m = "<strong>You are not authorized to access this resource.</strong >"
        RESPONSE.setStatus(401)
        RESPONSE.setHeader('WWW-Authenticate',
                           'basic realm=%s' % realm, 1 )
	raise Unauthorized, m

    def logout(self, REQUEST, RESPONSE):
	""" Allow user to invalidate authentication. """
	# To logout, invoke this method and cancel out of the
	# authentication process.

        user = self.getUser()
        if hasattr( user, 'getUserDN' ):
	    user._expire()
        realm = RESPONSE.realm
        RESPONSE.setStatus(401)
        RESPONSE.setHeader('WWW-Authenticate',
                           'basic realm=%s' % realm, 1 )
	return "You are logged out."

    def manage_configure( self, REQUEST=None):
        """
        Manage STLDAPManager properties.

        """
        message = ''
        if REQUEST.has_key('updateConfiguration'):
            if not REQUEST.has_key('cancel'):
                self.title = REQUEST.get( 'title', '' )

                ldapServer = REQUEST.get( 'ldapServer' )
                directoryBase = REQUEST.get( 'directoryBase' )
                personSearchBase = REQUEST.get( 'personSearchBase' )
                orgUnitBrowseBase = REQUEST.get( 'orgUnitBrowseBase' )
                groupBrowseBase = REQUEST.get( 'groupBrowseBase' )
                zopeRoleSearchBase = REQUEST.get( 'zopeRoleSearchBase' )
                zopeRoleSearchScope = REQUEST.get( 'zopeRoleSearchScope' )
                personNamingAttribute = REQUEST.get( 'personNamingAttribute' )
                personsOUAttr = REQUEST.get( 'personsOUAttr' )
                givenAttr = REQUEST.get( 'givenAttr' )
                groupObjectClass = REQUEST.get( 'groupObjectClass' )

                self.LDAPDirectory.setServers( ldapServer )
                self.LDAPDirectory.setDirectoryBase( directoryBase )

                if personSearchBase:
                    self.LDAPDirectory.setPersonSearchBase( personSearchBase )
                else:
                    self.LDAPDirectory.setPersonSearchBase( ldapDirectoryBase )

                if orgUnitBrowseBase:
                    self.LDAPDirectory.setOrgUnitBrowseBase( orgUnitBrowseBase )
                else:
                    self.LDAPDirectory.setOrgUnitBrowseBase( ldapDirectoryBase )

                if groupBrowseBase:
                    self.LDAPDirectory.setGroupBrowseBase( groupBrowseBase )
                else:
                    self.LDAPDirectory.setGroupBrowseBase( ldapDirectoryBase )

                if zopeRoleSearchBase:
                    self.LDAPDirectory.setZopeRoleSearchBase(zopeRoleSearchBase)
                else:
                    self.LDAPDirectory.setZopeRoleSearchBase( directoryBase )

                self.LDAPDirectory.setZopeRoleSearchScope(zopeRoleSearchScope)

                personSchema = self.LDAPDirectory._personManagedSchema
                if personNamingAttribute != '':
                    personSchema.setNamingAttribute(personNamingAttribute)
                if personsOUAttr == 'dn':
                    personSchema.setPersonsOUAttr(personsOUAttr)
                elif personsOUAttr == 'useGivenAttr':
                    personSchema.setPersonsOUAttr(givenAttr)

                groupSchema = self.LDAPDirectory._groupManagedSchema
                objectClassAttr = groupSchema.getAttributeSchema('objectClass')
                if groupObjectClass in ['groupOfNames','groupOfUniqueNames']:
                    objectClasses = ['top',groupObjectClass]
                    objectClassAttr.setRequiredValues( objectClasses )

                message = 'Updated configuration.'

        return self.manage_configureForm( self, REQUEST,
                                          manage_tabs_message=message)

    def manage_groups( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
        the 'group' objectCategory, after handling any update
        information that may have been in the request.
        """
        user = self.getUser()
        REQUEST.set( 'objectCategory', 'group' )
        zObject = ZObject( user, REQUEST, message )
        zObject = zObject.__of__( self )
        return zObject.showObjectForm()

    def manage_orgUnits( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
        the 'orgUnit' objectCategory, after handling any update
        information that may have been in the request.
        """
        user = self.getUser()
        REQUEST.set( 'objectCategory', 'orgUnit' )
        zObject = ZObject( user, REQUEST, message )
        zObject = zObject.__of__( self )
        return zObject.showObjectForm()

    def manage_people( self, REQUEST, message=None ):
        """
        Display the appropriate object management form for 
        the 'person' objectCategory, after handling any update
        information that may have been in the request.
        """
        user = self.getUser()
        REQUEST.set( 'objectCategory', 'person' )
        zObject = ZObject( user, REQUEST, message )
        zObject = zObject.__of__( self )
        return zObject.showObjectForm()

    def manage_reports( self, REQUEST, message=None ):
        """
        Display a form that allows the user to save configuration
        settings in a number of formats:  1) security settings in
        OpenLDAP slapd.conf format, 2) security settings as iPlanet
        ACI's in LDIF format, or 3) the entire configuration in
        SchemaDefaults format.
        """
        if not REQUEST.has_key( 'reportsFormID' ):
            REQUEST.set('reportsFormID','reportSchemaDoc')
        user = self.getUser()
        zReports = ZReports( user, REQUEST )
        zReports = zReports.__of__( self )
        return zReports.showReportsForm( message=message )

    def manage_roles( self, REQUEST, message=None ):
        """
        Display forms to configure the list of roles used to control
        access to information in the directory and in this object, and
        to assign this object's four permissions to roles.
        """
        user = self.getUser()
        zRoles = ZRoles( user, REQUEST )
        zRoles = zRoles.__of__( self )
        return zRoles.showRolesForm( message )

    def manage_schema( self, REQUEST, message=None ):
        """
        Display the appropriate schema management form, after 
        handling any update information that may have been in 
        the request.
        """
        if not REQUEST.has_key( 'schemaFormID' ):
            REQUEST.set('schemaFormID','schemaObjectClasses')
        user = self.getUser()
        zSchema = ZSchema( user, REQUEST )
        zSchema = zSchema.__of__( self )
        return zSchema.showSchemaForm( message )

    def manage_searching( self, REQUEST, message=None ):
        """
        Display the form for managing the public searching/org
        chart view of this object.
        """
        if not REQUEST.has_key( 'schemaFormID' ):
            REQUEST.set('schemaFormID','schemaSearchFilter')
        user = self.getUser()
        zSchema = ZSchema( user, REQUEST )
        zSchema = zSchema.__of__( self )
        return zSchema.showSchemaForm( message )

    def orgChart( self, REQUEST, message=None ):
        """
        Given a REQUEST containing a distinguished name, display it.
        Otherwise show the search form.

        """
        user = self.getUser()
        REQUEST.set('user',user)
        REQUEST.set( 'formID', 'orgChart' )
        return self.orgChartForm( self, REQUEST )

    def search( self, REQUEST, message=None ):
        """
        Show the search form.

        """
        user = self.getUser()
        REQUEST.set('user',user)
        REQUEST.set( 'formID', 'search' )
        return self.searchForm( self, REQUEST )

    def viewJpegPhoto( self, REQUEST, message=None ):
        """
        Display the data for a jpegPhoto attribute, in a way
        suitable for calling from an "img src=" tag.
        """
        user = self.getUser()
        REQUEST.set( 'objectCategory', 'person' )
        zObject = ZObject( user, REQUEST, message )
        zObject = zObject.__of__( self )
        return zObject.viewJpegPhoto()

    def getUser(self):
        """
        Return an AuthenticatedUser connection object in the case of
        an LDAP authenticated user.  A Zope user object created by
        one of Jens' products will have a 'distinguishedName' 
        attribute.  Otherwise, return an AnonymousUser connection.

        But because authenticated users can 
        change their own passwords, the Zope user object can be holding 
        on to invalid credentials.  If that is the case, then expire 
        the Zope user, and return an AnonymousUser connection.  Users
        not already logged in also get an AnonymousUser connection.
        """

        user = getSecurityManager().getUser()

        if hasattr( user, 'getUserDN' ):
            dn = user.getUserDN()
            pw = user._getPassword() 
            try:
                LDAPUser = AuthenticatedUser( dn, pw, self.LDAPDirectory,
                                              zopeUser=user)
            except INVALID_CREDENTIALS:
                user._expire()
                raise 'Unauthorized'
            except:
                LDAPUser = AnonymousUser( self.LDAPDirectory, zopeUser=user )
        else:            
            LDAPUser = AnonymousUser( self.LDAPDirectory, zopeUser=user )
        return LDAPUser

    def showSearchResults( self, objectQuery, sortKey, REQUEST,
                           objectCategory=None, searchBase=None, 
                           searchScope=None, filter=None, searchKeys=None ):
        """
        A flexible, public method for obtaining raw search results as a
        list of dictionaries, one for each entry.
        """
        user = self.getUser()
        return user.showSearchResults( objectQuery,
                           sortKey, objectCategory, searchBase, 
                           searchScope, filter, searchKeys )

def addForm( self, client=None, REQUEST={}, RESPONSE=None, 
             manage_tabs_message=None, **kw ):
    """
    Show the STLDAPManager add form, taking defaults from SchemaDefaults.

    """
    self.addFormSource = DTMLFile("dtml/addForm", globals(),
                                  form_title='Add ST LDAP Manager',
                                  help_product='STLDAPManager',
                                  help_topic='Add.stx')

    REQUEST.set( 'ldapServer', 
	REQUEST.get('ldapServer',defaultLDAPServer ))
    REQUEST.set( 'directoryBase', 
	REQUEST.get('directoryBase',defaultDirectoryBase ))
    REQUEST.set( 'personSearchBase', 
	REQUEST.get('personSearchBase',defaultPersonSearchBase ))
    REQUEST.set( 'orgUnitBrowseBase', 
	REQUEST.get('orgUnitBrowseBase',defaultOrgUnitBrowseBase ))
    REQUEST.set( 'groupBrowseBase', 
	REQUEST.get('groupBrowseBase',defaultGroupBrowseBase ))
    REQUEST.set( 'zopeRoleSearchBase', 
	REQUEST.get('zopeRoleSearchBase',defaultZopeRoleSearchBase ))
    REQUEST.set( 'zopeRoleSearchScope', 
	REQUEST.get('zopeRoleSearchScope',defaultZopeRoleSearchScope ))
    REQUEST.set( 'personNamingAttribute', 
	REQUEST.get('personNamingAttribute',defaultUserNamingAttribute ))
    REQUEST.set( 'personsOUAttr', 
	REQUEST.get('personsOUAttr',defaultPersonsOUAttr ))
    REQUEST.set( 'groupObjectClass', 
	REQUEST.get('groupObjectClass',defaultGroupObjectClass ))
    return self.addFormSource( self, REQUEST,
                               manage_tabs_message=manage_tabs_message )

def addFunction( dispatcher, id, title, ldapServer, directoryBase, 
                 personSearchBase, orgUnitBrowseBase, groupBrowseBase,
                 zopeRoleSearchBase, zopeRoleSearchScope, 
                 personNamingAttribute, personsOUAttr, givenAttr,
                 groupObjectClass, REQUEST=None):
    """ 
    Create a new STLDAPManager and add it to myself.

    """
    # Trap for an empty id or ldapServer.  Redraw the form with an 
    # error message.
    #
    if str(id) == '' and REQUEST is not None:
        message = "You must supply an id for this object."
        return dispatcher.Control_Panel.Products.STLDAPManager.addForm( 
                        dispatcher, REQUEST, manage_tabs_message=message )

    if  not ldapServer and REQUEST is not None:
        message = """You must supply an LDAP Server.  Zope must be able
                     to connect to the LDAP server to properly initialize
                     this object."""
        return dispatcher.Control_Panel.Products.STLDAPManager.addForm( 
                        dispatcher, REQUEST, manage_tabs_message=message )
    
    inst = STLDAPManager( id, title, ldapServer, directoryBase, 
                          personSearchBase, orgUnitBrowseBase, 
                          groupBrowseBase, zopeRoleSearchBase, 
                          zopeRoleSearchScope, personNamingAttribute,
                          personsOUAttr, givenAttr, groupObjectClass,
                          REQUEST )
    dispatcher.Destination()._setObject(id, inst)

    if REQUEST is not None:
        dispatcher.manage_main(dispatcher, REQUEST)

InitializeClass(STLDAPManager)
