"""

    SearchLister.py
    $Author: payne $
    $Date: 2003/09/13 14:49:59 $
    
Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    The SearchLister class encapsulates a transient object that 
    takes a search string, converts it to an LDAP filter on the
    configured searchKey attributes, performs the search, and
    returns the results as a sorted list of dictionaries.

"""
__version__ = '$Revision: 1.13 $'[11:-2]
# $Source: /Users/payne/job/ldap/CVS/STLDAPManager/SearchLister.py,v $

import re
from string import find, join, lower

from LDAPExceptions import *
from LDAPUtilities import _ldap, explode, wildDN

ws = re.compile( '\s+' )

class SearchLister:

    def __init__(self, user, objectQuery, sortKey, objectCategory, 
                 searchBase, searchScope, filter, searchKeys, schema, ):
        # sortKey becomes an attribute, as the only way to pass it to
        # the sortFunction.

        if lower(objectCategory) == 'person':
            schema = user.LDAPDirectory.getPersonSchema()
        elif lower(objectCategory) == 'orgunit':
            schema = user.LDAPDirectory.getOrgUnitSchema()
        elif lower(objectCategory) == 'group':
            schema = user.LDAPDirectory.getGroupSchema()

        self.sortKey = sortKey

        namingAttribute = schema.getNamingAttribute()
        if lower(objectCategory) == 'person':
            personsOUAttr = schema.getPersonsOUAttr()

        orgUnitNamingAttribute = user.LDAPDirectory.getOrgUnitSchema().getNamingAttribute()

        if searchBase is None:
            if lower(objectCategory) == 'person':
                searchBase = user.LDAPDirectory.getPersonSearchBase()
            elif lower(objectCategory) == 'orgunit':
                searchBase = user.LDAPDirectory.getOrgUnitBrowseBase()
            elif lower(objectCategory) == 'group':
                searchBase = user.LDAPDirectory.getGroupBrowseBase()

        if searchScope is None:
            searchScope = _ldap.SCOPE_SUBTREE

        if searchKeys is None:
            searchKeys = user.getReadableSearchKeys(schema)
            
        # Build the search filter.  Check the special cases first:
        #
        #   1. objectCategory is 'person' and objectQuery starts
        #      with 'personNamingAttribute='.  This means look up
        #      just this user.
        #
        #   2. objectCategory is 'person', 'personsOUAttr' is not 'dn'
        #      and objectQuery starts with 'personsOUAttr='.  This means 
        #      search for all of the people in this OU, in the case
        #      where people's OU is determined by an attribute.
        #
        #   3. objectCategory is 'person', 'personsOUAttr' is 'dn',
        #      and the query looks like a distinguishedName.  This
        #      means search for all of the people in the OU specified
        #      by the DN, and corresponds to the case where people's
        #      OU is determined by looking at their DN.
        #
        #   4. objectCategory is not person, and the objectQuery
        #      the objectQuery starts with the naming attribute for
        #      that objectCategory, and an equals sign.  Return the
        #      target object and sub-objects of this category.

        # Split the query on white space.

        if objectQuery is not None:
            queryWords = ws.split( objectQuery )
        if filter is None:

            #   ... then we'll have to figure something out.  If the
            #   query looks like a distinguished name, then try out
            #   the options described above.

            filter = '(&' + schema.getSearchFilter()
            dnQuery = 0
            if find( objectQuery, '=' ) != -1:
                normalizedQuery = lower( join(explode(objectQuery),',') )
                if objectCategory == 'person' and \
                 find( normalizedQuery, '%s=' % lower(namingAttribute) ) == 0:
                    filter = filter + \
                             '(' + objectQuery + '))'
                    dnQuery = 1
                elif objectCategory == 'person' and \
                 personsOUAttr != 'dn' and \
                 find( normalizedQuery, '%s=' % lower(personsOUAttr) ) == 0:
                    filter = filter + \
                         '(%s=*' % personsOUAttr + wildDN(objectQuery) + '))'
                    dnQuery = 1
                elif objectCategory == 'person' and \
                 personsOUAttr == 'dn' and \
                 find( normalizedQuery, '%s=' % lower(orgUnitNamingAttribute)) == 0:
                    filter = schema.getSearchFilter()
                    searchBase = normalizedQuery 
                    dnQuery = 1
                elif objectCategory != 'person' and \
                 find( normalizedQuery, '%s=' % lower(namingAttribute) ) == 0:
                    filter = schema.getSearchFilter()
                    searchBase = normalizedQuery 
                    dnQuery = 1

            #   Otherwise do the standard thing of looking for each of
            #   the words in the query in each of the attributes being
            #   searched.

            if not dnQuery:
                filter = filter + '(|'
                for word in queryWords:
                    for item in searchKeys:
                        queryTerm = '(' + item + '=*' + word + '*)'
                        filter = filter + queryTerm
                filter = filter + '))'

        # Search!

        try:
            results = user.search(searchBase, searchScope, filter, 
                                  attrList=searchKeys)
        except NO_SUCH_OBJECT:
            self.resultsList = []
            return

        # Search results is a list of dictionaries of dictionaries, with
        # distiguished name as the key.  Put the distinguished name into
        # the dictionary and reduce results to a list of dictionaries.
        # Add a visible '-' in place of missing attributes.

        self.resultsList = len(results)*['']
        for i in range(len(results)):
            item = results[i]
            dn = item[0]
            attrs = item[1]
            attrs['distinguishedName'] = [dn]
            for key in searchKeys:
                if not attrs.has_key(key):
                    attrs[key] = '-'
            self.resultsList[i] = attrs

        # Apply the default sorting.

        self.resultsList.sort(self.__sortFunction)

    def __sortFunction( self, x, y ):
        """
            Helper function to sort a list of dictionaries on an
            arbitrary dictionary key.  The dictionary items are
            assumed to be sequences and the sort is done on the
            first item in the sequence.
        """
        if x.has_key(self.sortKey):
            xval = x[self.sortKey][0]
        else:
            xval = ''

        if y.has_key(self.sortKey):
            yval = y[self.sortKey][0]
        else:
            yval = ''

        return cmp(lower(xval), lower(yval))
