"""

    ZLDAPTree.py
    $Author: payne $
    $Date: 2003/12/07 04:13:47 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    A class for rendering a tree of LDAP objects.  A search returns
    a list of Python search results (tuple of DN followed by a
    dictionary of attribute values) representing a tree of objects.
    This result is filtered into a dictionary keyed on the DN of
    a node in the tree, where the value is a list of MappingObjects
    that represent the leaves of that node.  This results in an
    efficient method for building the tree.

"""

__version__ = '$Revision: 1.9 $'[11:-2]

from copy import deepcopy
from DocumentTemplate.DocumentTemplate import HTML
from LDAPExceptions import INVALID_DN_SYNTAX, NO_SUCH_OBJECT
from LDAPUtilities import _ldap
from LDAPUtilities import *
from SearchLister import SearchLister
from string import join, lower, split, strip
from urllib import quote, quote_plus
import sys, types

class ZLDAPTree:

    def getChildNodes( self, mappingDN, nodeDict ):
        """
        Return the list of MappingObjects stored under the key mappingDN.
        """
        return nodeDict[ mappingDN ]

    def getNodeDict( self, user, base, schema,
                     filter='(objectClass=organizationalUnit)' ):
        """
        Perform the filtered query for objects in a tree under searchBase.  
        Build dictionary of objects under each node.  Save the base under the
        'base' key.
        """
        base = join( explode(base), ',' )
        browseKeys = schema.getBrowseKeys()
        if browseKeys:
            attrs = browseKeys
            format = schema.getBrowseFormat()
        else:
            attrs = ['objectClass']
            format = '%s'

        result = user.search( base,_ldap.SCOPE_SUBTREE, filter, attrs )

        nodeDict = {}
        name = explode( base,notypes=1 )[0]
        nodeDict[ 'base' ] = MappingObject( base, name )

        for item in result:
            #
            # To find the key, explode each dn into its components, drop
            # the first, and join together again.  For consistency, do
            # the same for each MappingObject's dn.

            exploded = explode(item[0])
            key = join( exploded[1:], ',' )
            dn = join( exploded, ',' )
            if browseKeys:
                name = format % \
                       tuple(map(lambda i,item=item: item[1][i][0], browseKeys))
            else:
                name = explode( exploded[0],notypes=1 )[0]
            if nodeDict.has_key( key ):
                nodeDict[ key ].append( MappingObject( dn, name ) )
            else:
                nodeDict[ key ] = [ MappingObject( dn, name ) ]
            if not nodeDict.has_key( dn ):
                nodeDict[ dn ] = []

        return nodeDict

    def getTreeCancelHiddenTags(self, REQUEST):
        """
        Build hidden tree tags for the form to cancel out of a selection
        form with an ldapTree on it.
        """
        reqForm = self.getTreeQueryDict(REQUEST)
        if reqForm.has_key('updateObject'):
            del reqForm['updateObject']

        reqForm['updateObject'] = 1

        output = ''
        for key in reqForm.keys():
            if key[0:9] == "selectDN_":
                key = key[9:]
                values = REQUEST.get(key,[])
                if type(values)==type(''):
                    values = [ values, ]
                for value in values:
                    output = output + """
        <input type="hidden" name="%s" value="%s">""" % \
                           ( key, str( value ) )
            elif type( reqForm[key] ) == type([]):
                for value in reqForm[key]:
                    output = output + """
        <input type="hidden" name="%s" value="%s">""" % \
                           ( key, str( value ) )
            else:
                output = output + """
        <input type="hidden" name="%s" value="%s">""" % \
                           ( key, str( reqForm[key] ) )

        return output

    def getTreeQuery( self, REQUEST ):
        # Build a query string to be used as the urlparam of a
        # tree tag.
        #
        reqForm = self.getTreeQueryDict( REQUEST )

        targetAttribute = REQUEST.get('targetAttribute','')
        if not reqForm.has_key( 'selectDN_'+targetAttribute ):
            reqForm['selectDN_'+targetAttribute] =  ''

        return self.urlencode( reqForm, 1 )

    def getTreeQueryDict( self, REQUEST ):
        #
        # Build a cleaned-up version of REQUEST.form containing only
        # the items necessary for building the tree
        #
        reqForm = deepcopy(REQUEST.form)
        if reqForm.has_key('queryString'):
            del reqForm['queryString']
        if reqForm.has_key('tree-s'):
            del reqForm['tree-s']
        if reqForm.has_key('tree-e'):
            del reqForm['tree-e']
        if reqForm.has_key('tree-c'):
            del reqForm['tree-c']

        targetAttribute = REQUEST.get('targetAttribute','')
        if reqForm.has_key(targetAttribute):
            del reqForm[targetAttribute]

        return reqForm

    def getTreeQueryReturn( self, REQUEST, returnsToFormID='display' ):
        """
        Generate a tree query string suitable for returning to the
        calling form.
        """
        reqForm = deepcopy(self.getTreeQueryDict( REQUEST ))

        targetAttribute = REQUEST.get('targetAttribute','')
        if reqForm.has_key( 'selectDN_'+targetAttribute ):
            del reqForm['selectDN_'+targetAttribute]

        reqForm['objectFormID'] = returnsToFormID

        return '?' + self.urlencode( reqForm, 1 ) + '&'

    def ldapTree(self, user, REQUEST,
                  base=None, 
                  query='(objectClass=organizationalUnit)', 
                  urlparam='formID=orgChart',
                  header=None,
                  leaves=None,
                  branchMethod="getChildNodes(mappingDN,nodeDict)",
                  showBranchesAsLinks=1,
                  action="&dtml-URL1;/manage_orgUnits?",
                  queryStringReturn='',
                  targetAttribute='distinguishedName',
                  returnsToFormID='display',
                  linksTo='mappingDN',
                  namingAttribute='ou',
                  cancel=None,
                  none=None,
                  objectCategory='orgUnit' ):
        """
        Generate a tree view of the directory nodes encoded in the
        nodeDict dictionary.
        """
        REQUEST.set( 'user', user )
        if base is None:
            base = user.LDAPDirectory.getOrgUnitBrowseBase()
        REQUEST.set( 'mappingDN', base )

        if objectCategory == 'group':
            nodeSchema = user.getLDAPDirectory().getGroupSchema()
        elif objectCategory == 'person':
            nodeSchema = user.getLDAPDirectory().getPersonSchema()
        elif objectCategory == 'orgUnit':
            nodeSchema = user.getLDAPDirectory().getOrgUnitSchema()
        REQUEST.set( 'nodeSchema',nodeSchema)

        browseKeys = nodeSchema.getBrowseKeys()

        nodeDict = self.getNodeDict( user, base, nodeSchema, filter=query )
        REQUEST.set( 'nodeDict', nodeDict )

        REQUEST.set( 'showBranchesAsLinks', showBranchesAsLinks )

        if header is not None:
            headerStr = 'header="%s"\n' % header
        else:
            headerStr = ''

        if leaves is not None:
            leavesStr = 'leaves="%s"\n' % leaves
        else:
            leavesStr = ''

        if urlparam == 'getTreeQuery':
            urlparam = self.getTreeQuery( REQUEST )

        if queryStringReturn == 'getTreeQueryReturn':
            queryStringReturn = \
                self.getTreeQueryReturn( REQUEST, 
                                         returnsToFormID=returnsToFormID )

        if linksTo == 'mappingName':
            mappingLink = \
                """%s=<dtml-var "user.getNameFromDN(mappingDN)" url_quote>""" \
                                        % ( namingAttribute )
        else:
            mappingLink = "<dtml-var mappingDN url_quote>"

        if none is not None:
            none = """
            <input type="submit" name="selectDN_%s" value=" None ">""" % \
                                       ( targetAttribute )
        else:
            none = ''

        # Here's a string template that holds the DTML returned by
        # this method.  Notice the replacement strings at the end.
        #
        treeString = """
<dtml-with "REQUEST['user']">
<dtml-call "REQUEST.set('nodeDict', getNodeDict(user,mappingDN,nodeSchema,filter='%s'))">
<table border="0" cellspacing="0">
  <tr style="white-space: nowrap">

    <td width="16" nowrap>&nbsp;</td>
<dtml-if showBranchesAsLinks>
    <td colspan="2" style="text-decoration: none;">
      <a style="text-decoration: none;" href="%s%s%s=%s">
        <dtml-var "nodeDict['base'].mappingName"></a>
<dtml-else>
    <td colspan="2" style="font-family: Verdana, Helvetica, sans-serif;text-decoration: none;">
        <dtml-var "nodeDict['base'].mappingName">
</dtml-if>
    </td>
  </tr>
  <tr>
    <td width="16" nowrap>&nbsp;</td>
    <td width="16">&nbsp;&nbsp;</td>
    <td style="font-family: Verdana, Helvetica, sans-serif;">
    <dtml-tree urlparam="%s"
               id="mappingDN"
               sort="mappingName"
               %s%sbranches_expr="%s">
    <dtml-if showBranchesAsLinks>
      <a style="text-decoration: none;" href="%s%s%s=%s">&dtml-mappingName;</a>
    <dtml-else>
      &dtml-mappingName;
    </dtml-if>
    </dtml-tree>
  </dtml-with user>
  </td>
</tr>""" %                 ( 
                             query,
                             action,
                             queryStringReturn,
                             targetAttribute,
                             mappingLink,
                             urlparam,
                             headerStr,
                             leavesStr,
                             branchMethod,
                             action,
                             queryStringReturn,
                             targetAttribute,
                             mappingLink,
                           )

        if cancel is not None or none is not None:
            treeString = treeString + """
      <form action="&dtml-URL0;" method="POST">

      %s

        <tr>
          <td width="16">&nbsp;</td>
        </tr>

        <tr>
          <td width="16">&nbsp;</td>
          <td colspan="2" class="form-element">
            <input type="submit" name="treeCancel" value=" Cancel ">%s
          </td>
        </tr>

      </form>""" % ( 
                     self.getTreeCancelHiddenTags( REQUEST ),
                     none )

        treeString = treeString + """
      </table>"""

        #return treeString

        # Make a DocumentTemplate.HTML object out of the string, and
        # render it.
        #
        ob = HTML( treeString )
        return apply(HTML.__call__, ( ob, self, REQUEST ))

    def orgUnitMemberBrowser(self,junk,REQUEST,**args):
        """
        Display org chart search results:
        """
        user = REQUEST['user']
        level = REQUEST['tree-level']
        colspan = REQUEST['tree-colspan']
        dataspan = colspan - level
        if dataspan == 2:
        
            mappingDN = REQUEST['mappingDN']
            schema = user.LDAPDirectory.getPersonSchema()
            browseKeys = schema.getBrowseKeys()
            browseFormat = schema.getBrowseFormat()
            sortKey = schema.getBrowseSortKey()
            namingAttribute = schema.getNamingAttribute()
            if namingAttribute in browseKeys:
                searchKeys = deepcopy(browseKeys)
            else:
                searchKeys = deepcopy(browseKeys)+[namingAttribute]
            try:
                if schema.getPersonsOUAttr() == 'dn':
                    filter = "(&%s)" % schema.getSearchFilter()
                    results = SearchLister( user, None, sortKey, 'person',
                                            mappingDN, _ldap.SCOPE_ONELEVEL,
                                            filter, searchKeys, 
                                            schema ).resultsList
                else:
                    searchBase = user.LDAPDirectory.getPersonSearchBase()
                    filter = "(&%s(%s=%s))" % ( schema.getSearchFilter(),
                                                schema.getPersonsOUAttr(),
                                                mappingDN )
                    results = SearchLister( user, None, sortKey, 'person',
                                            searchBase, _ldap.SCOPE_SUBTREE,
                                            filter, searchKeys,
                                            schema ).resultsList
            except NO_SUCH_OBJECT:
                return ''

            _td_colspan='<TD COLSPAN="%s" NOWRAP></TD>'
            _td_single ='<TD WIDTH="16" NOWRAP></TD>'
            if level > 1:   h=_td_colspan % (level)
            elif level > 0: h=_td_single  * (level)
            else: h=''

            output = args['standard_html_header']
            if REQUEST.has_key('private'):
                for row in results:
                    output = output + """<a style="text-decoration: none"
href="%s?distinguishedName=%s">
&nbsp;&nbsp;""" %           ( REQUEST['URL0'],
                              quote_plus(row['distinguishedName'][0]), )
                    if browseKeys:
                        output = output + browseFormat % \
                            tuple(map(lambda i,row=row: row[i][0], browseKeys))
                    else:
                        output = output + row[namingAttribute][0]
                    output = output + "</a><br>"
            else:
                for row in results:
                    output = output + """<a style="text-decoration: none"
href="%s?""" %              ( REQUEST['URL0'] )
                    if REQUEST.has_key('extraQuery'):
                        output = output + "%s&" % ( REQUEST['extraQuery'] )
                    output = output + """query=%s%%3D%s">
&nbsp;&nbsp;""" %           ( namingAttribute,
                              row[namingAttribute][0] )
                    if browseKeys:
                        output = output + browseFormat % \
                            tuple(map(lambda i,row=row: row[i][0], browseKeys))
                    else:
                        output = output + row[namingAttribute][0]
                    output = output + "</a><br>"
            output = output + args['standard_html_footer']
            return output
        else:
            return ''

    # urlencode needs the doseq option that came after Python1.5.2
    # This code comes from Python 2.1.3's urllib.py

    def urlencode(self,query,doseq=0):
        """
        Encode a sequence of two-element tuples or dictionary into a URL 
        query string.
    
        If any values in the query arg are sequences and doseq is true, each
        sequence element is converted to a separate parameter.
    
        If the query arg is a sequence of two-element tuples, the order of the
        parameters in the output will match the order of parameters in the
        input.
        """
    
        if hasattr(query,"items"):
            # mapping objects
            query = query.items()
        else:
            # it's a bother at times that strings and string-like objects are
            # sequences...
            try:
                # non-sequence items should not work with len()
                x = len(query)
                # non-empty strings will fail this
                if len(query) and type(query[0]) != types.TupleType:
                    raise TypeError
                # zero-length sequences of all types will get here and succeed,
                # but that's a minor nit - since the original implementation
                # allowed empty dicts that type of behavior probably should be
                # preserved for consistency
            except TypeError:
                ty,va,tb = sys.exc_info()
                raise TypeError, "not a valid non-string sequence or mapping object", tb
    
        l = []
        if not doseq:
            # preserve old behavior
            for k, v in query:
                k = quote_plus(str(k))
                v = quote_plus(str(v))
                l.append(k + '=' + v)
        else:
            for k, v in query:
                k = quote_plus(str(k))
                if type(v) == types.StringType:
                    v = quote_plus(v)
                    l.append(k + '=' + v)
#                elif type(v) == types.UnicodeType:
#                    # is there a reasonable way to convert to ASCII?
#                    # encode generates a string, but "replace" or "ignore"
#                    # lose information and "strict" can raise UnicodeError
#                    v = quote_plus(v.encode("ASCII","replace"))
#                    l.append(k + '=' + v)
                else:
                    try:
                        # is this a sufficient test for sequence-ness?
                        x = len(v)
                    except TypeError:
                        # not a sequence
                        v = quote_plus(str(v))
                        l.append(k + '=' + v)
                    else:
                        # loop over the sequence
                        for elt in v:
                            l.append(k + '=' + quote_plus(str(elt)))
        return join(l,'&')

class MappingObject:
    """
    This object is defined as a convenience.  Having getChildNodes 
    return a list of MappingObjects instead of a list of dictionaries
    makes it easier to use the Zope tree tag.      
    """
    def __init__(self, mappingDN, mappingName):
        self.mappingDN = mappingDN
        self.mappingName = mappingName

    def __repr__(self):
        return '[' + self.mappingDN + ': ' + self.mappingName + ']'   

