"""
    ZObject.py
    Author: $Author: payne $
    Date:   $Date: 2003/12/13 04:50:33 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

    The ZObject class is a Zope-aware object that instantiates
    an internal LDAP object appropriate for the DN information
    passed in the REQUEST.  It contains methods used for drawing
    the forms to display an LDAP object's attributes, and for
    updating the LDAP object from REQUEST information submitted
    from forms.
"""

__version__ = '$Revision: 1.71 $'[11:-2]

from AccessControl import getSecurityManager
from Acquisition import Implicit
import base64
from copy import deepcopy
from DocumentTemplate.DocumentTemplate import HTML
from Globals import DTMLFile, InitializeClass
from os import fstat
from string import join, lower, split
import sys 
from time import localtime
import types
from urllib import quote_plus, unquote_plus
from zExceptions import Unauthorized

from LDAPExceptions import *
from LDAPObject import LDAPObject
from LDAPUtilities import _ldap, DNNormalize, DNsAreEquivalent, \
    getNameFromDN, getParentFromDN, listsAreEquivalent, list1ContainsList2, \
    mergeLists, TRUE, FALSE

class ZObject( Implicit ):

    # Set a limit, in bytes, on how big a photo can be uploaded.  Set to
    # zero if you don't want a limit

    maxJpegPhotoSize = 100000

    #   DTML templates for the various "faces" of the Objects tab.
    #   Each template's forms must contain a hidden variable that
    #   identifies the form.  The name is 'objectFormID', and the
    #   value must match one of the tests for it in _manage_objects.
    #
    #     <input type=hidden name=objectFormID value=objectAdd>

    #   DTML templates for the various "faces" of the Search tab.

    manage_objectFormSelection = DTMLFile('dtml/objectFormSelection', globals())
    _browseGroupsForm = DTMLFile('dtml/objectBrowseGroups', globals())
    _browseOrgUnitsForm = DTMLFile('dtml/objectBrowseOrgUnits', globals())
    _browsePeopleForm = DTMLFile('dtml/objectBrowsePeople', globals())
    _displayForm = DTMLFile('dtml/objectDisplay', globals())
    _objectAddForm = DTMLFile('dtml/objectAdd', globals())
    _objectDelForm = DTMLFile('dtml/objectDel', globals())
    _objectGroupMembershipForm = DTMLFile('dtml/objectGroupMembership', globals())
    _objectRenameForm = DTMLFile('dtml/objectRename', globals())
    _objectDelPersonForm = DTMLFile('dtml/objectDelPerson', globals())
    _objectRenamePersonForm = DTMLFile('dtml/objectRenamePerson', globals())
    _renamePersonForm = DTMLFile('dtml/renamePerson', globals())
    _searchForm = DTMLFile('dtml/objectSearch', globals(),
                           management_view='People')
    _selectGroupDNForm = DTMLFile('dtml/selectGroupDN', globals())
    _selectOrgUnitDNForm = DTMLFile('dtml/selectOrgUnitDN', globals())
    _selectGroupByBrowsingForm = DTMLFile('dtml/selectGroupByBrowsing',
                           globals(), management_view='Groups')
    _selectOrgUnitByBrowsingForm = DTMLFile('dtml/selectOrgUnitByBrowsing',
                           globals(), management_view='Org Units')
    _selectPersonByBrowsingForm = DTMLFile('dtml/selectPersonByBrowsing',
                           globals(), management_view='People')
    _selectPersonDNForm = DTMLFile('dtml/selectPersonDN', globals())
    _selectPersonOrgUnitDNForm = DTMLFile('dtml/selectPersonOrgUnitDN',globals())
    _updateJpegPhotoForm = DTMLFile('dtml/updateJpegPhoto', globals())

    def __init__( self, user, REQUEST, message=None ):

        self.user = user
        self.message = message
        self.request = REQUEST

        #   Get an LDAPObject.  

        self.object = LDAPObject( user )
        distinguishedName = self.request.get( 'distinguishedName', None )

        self._IChangedMyOwnPassword = FALSE

        if distinguishedName is not None:
            distinguishedName = unquote_plus( distinguishedName )
            self.object.refresh( distinguishedName )
            self.objectCategory = self.object.getObjectCategory()
        elif self.request.has_key( 'objectCategory' ):
            self.objectCategory = self.request['objectCategory']
        else:
            self.objectCategory = 'person'

        if lower(self.objectCategory) == 'person':
            self.schema = user.LDAPDirectory.getPersonSchema()
        elif lower(self.objectCategory) == 'orgunit':
            self.schema = user.LDAPDirectory.getOrgUnitSchema()
        elif lower(self.objectCategory) == 'group':
            self.schema = user.LDAPDirectory.getGroupSchema()
        else:
            raise UNRECOGNIZED_OBJECT

        self.namingAttribute = self.schema.getNamingAttribute()
        namingAttributeSchema = \
            self.schema.getAttributeSchema(self.namingAttribute)
        self.friendlyNamingAttribute = namingAttributeSchema.getFriendlyName()

    def showObjectForm(self):

        #   Put into the REQUEST information needed on the forms
       
        message = self.message

        #   If we got here by a form submit button, then we need to
        #   update the directory first.  Any form that needs to hit
        #   the methods in ZObject.updateDirectoryFromRequest must
        #   contain:
        #
        #   <input type="hidden" name="updateObject" value="1">

        if self.request.has_key('updateObject'):
            if not self.request.has_key('cancel'):
                self.message = self.updateDirectoryFromRequest()

                if self.iChangedMyOwnPassword():
                    #   ... then I can't go on.  Expire the Zope user 
                    #   object, and redirect to the search screen,  
                    #    where I can login again.

                    getSecurityManager().getUser()._expire()
                    url = self.request.get('URL1') + '/'
                    self.request.RESPONSE.redirect(url)

            else:

                if self.request.has_key('returnsToFormID') \
                 and self.request['returnsToFormID'] != self.request['objectFormID']:
                    self.request['objectFormID'] = \
                        self.request['returnsToFormID']

                else:
                    #   Clean up the REQUEST object by just doing a
                    #   redirect back to this form for this object.
                    url = "%s?objectFormID=%s" % \
                          ( self.request['URL0'],
                            self.request['objectFormID'] )
                    if self.request.has_key('distinguishedName'):
                        url = url + "&distinguishedName=%s" % \
                          ( quote_plus(self.request['distinguishedName']) )
                    if self.request.has_key('displayCategory'):
                        url = url + "&displayCategory=%s" % \
                          ( quote_plus(self.request['displayCategory']) )
                    self.request.RESPONSE.redirect( url )

        #   Set default form, and render it.  Each form must contain a
        #   hidden tag called 'schemaFormID' whose value, with '_'
        #   prepended, and 'Form' appended, is form rendering DTMLFile
        #   call.
    
        if not self.request.has_key('objectFormID'):
            if self.request.has_key('distinguishedName'):
                self.request.set('objectFormID','display')
            else:
                self.request.set('objectFormID','search')

        if self.request.objectFormID == 'display':
            return self._displayForm( 
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'addObject':
            return self._objectAddForm( 
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'delObject':
            if self.request['objectCategory'] == 'person':
                return self._objectDelPersonForm(
                        self, self.request, manage_tabs_message = self.message )
            else:
                return self._objectDelForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'renameObject':
            if self.request['objectCategory'] == 'person':
                return self._objectRenamePersonForm(
                        self, self.request, manage_tabs_message = self.message )
            else:
                return self._objectRenameForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'renamePerson':
            return self._renamePersonForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'groupMembership':
            return self._objectGroupMembershipForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectPersonOrgUnitDN':
            return self._selectPersonOrgUnitDNForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectOrgUnitDN':
            return self._selectOrgUnitDNForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectGroupDN':
            return self._selectGroupDNForm(
                        self, self.request, manage_tabs_message = self.message )

#        elif self.request.objectFormID == 'browseOrgUnits':
#            return self._browseOrgUnitsForm(
#                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'browseGroups':
            return self._browseGroupsForm(
                        self, self.request, manage_tabs_message = self.message )

        elif self.request.objectFormID == 'selectPersonDN':
            return self._selectPersonDNForm( self, self.request )

        elif self.request.objectFormID == 'selectOrgUnitDN':
            return self._selectOrgUnitDNForm( self, self.request )

        elif self.request.objectFormID == 'selectGroupDN':
            return self._selectGroupDNForm( self, self.request )

        elif self.request.objectFormID == 'selectGroupByBrowsing':
            return self._selectGroupByBrowsingForm( self, self.request )

        elif self.request.objectFormID == 'selectOrgUnitByBrowsing':
            return self._selectOrgUnitByBrowsingForm( self, self.request )

        elif self.request.objectFormID == 'selectPersonByBrowsing':
            return self._selectPersonByBrowsingForm( self, self.request )

        elif self.request.objectFormID == 'search':
            return self._searchForm( self, self.request, )

        elif self.request.objectFormID == 'orgChart':
            if self.request.objectCategory == 'person':
                return self._browsePeopleForm( self, self.request )
            elif self.request.objectCategory == 'orgUnit':
                return self._browseOrgUnitsForm( self, self.request )
            elif self.request.objectCategory == 'group':
                return self._browseGroupsForm( self, self.request )

	elif self.request.objectFormID == 'updateJpegPhoto':
            return self._updateJpegPhotoForm( self, self.request, )

        else:
            self.message = """
                      The previous form's objectFormID does not match a
                      known form display method.
                      """
            self.request.set('objectFormID','search')
            return self._searchForm(
                self, self.request, manage_tabs_message = self.message)


    def canControl( self, objectCategory ):
        """
            Can the user be a controller for objects of the specified
            objectCategory?  That is, does the user have any roles
            in the ManagedSchema's controllers list.
        """
        controllers = self.schema.getControllers()
	userRoles = self.user.getRoles()
        for item in userRoles:
            if item in controllers:
                return TRUE
        return FALSE

    def getGroupMembers( self ):
        """
            If the saved LDAPObject has been refreshed with information
            from a groupOfNames or groupOfUniqueNames in the directory,
            then return the list of values (DNs) of the member and/or
            uniqueMember attributes.
        """
        groupMembers = []
        objectClassAttr = self.object.getManagedAttribute('objectClass')
        if 'groupOfNames' in objectClassAttr.getValidValues():
            groupMembers.extend(self.object.getManagedAttribute('member').getValues())
        if 'groupOfUniqueNames' in objectClassAttr.getValidValues():
            groupMembers.extend( self.object.getManagedAttribute('uniqueMember').getValues())
        groupMembers.sort()
        return groupMembers

    def getObject( self ):
        """
            Returns the saved LDAPObject object
        """
        return self.object

    def getObjectClasses( self ):
        """
            Returns the values of the objectClass attribute for the
            saved LDAPObject
        """
        return self._objectClasses

    def getReadableSearchKeys(self):
        """
            Returns a filtered list of search keys, containing only
            those readable by the user.
        """
        searchKeys = deepcopy( self.schema._searchKeys )
        for attributeName in searchKeys:
            attributeSchema = self.schema.getAttributeSchema(attributeName)
            readers = attributeSchema.getReaders()
            canRead = FALSE
            for role in self.user.getRoles():
                if role in readers:
                    canRead = TRUE
                    break
            if not canRead:
                searchKeys.remove(attributeName)
        return searchKeys

    def iChangedMyOwnPassword( self ):
        """
            Returns the value of the attribute that checkes whether 
            I changed my own password.
        """
        return self._IChangedMyOwnPassword

    def selectDNKey(self):
        """
            If the REQUEST contains a key of the form 'selectDN_attrName'
            this method returns 'attrName'
        """
        for key in self.request.keys():
            if key[0:9] == 'selectDN_':
                return key[9:]
        return ''

    def uniqueID(self):
        """
	Generate an ID "unique" to a request by concatenating the
        user's DN with a time stamp and taking the md5 checksum
        of that.  Append this to img tags to evade browser caching
	problems.
        """
        return base64.encodestring( self.user.distinguishedName + str(localtime()) )

    def updateDirectoryFromRequest(self):
        """
            Make directory updates according to the REQUEST variables.
        """

        #   Invoke a method that matches the name of the submit
        #   button (with '_' prepended).

        if self.request.has_key( 'updateObjectValues' ):
            message = self._updateObjectValues()

        elif self.request.has_key( 'updateMembership' ):
            url = self.request['URL1'] + \
                  "/manage_groups?distinguishedName=" + \
                  "%s" % quote_plus(self.request.distinguishedName) + \
                  "&objectFormID=groupMembership"
            self.request.RESPONSE.redirect( url )
            message = url

        elif self.request.has_key( 'addObject' ):
            message = self._addObject()

        elif self.request.has_key( 'delObject' ):
            message = self._delObject()

        elif self.request.has_key( 'renameObject' ):
            if self.request['objectCategory'] == 'person':
                message = self._renamePerson()
            else:
                message = self._renameObject()

        elif self.request.has_key( 'addMembers' ):
            message = self._addMembers()

        elif self.request.has_key( 'removeMembers' ):
            message = self._removeMembers()

        elif self.request.has_key( 'selectObjectByBrowsing' ):
            message = self._selectObjectByBrowsing()

	elif self.request.has_key( 'treeCancel' ):
	    if self.request.has_key('treeCancel'):
                self.request.RESPONSE.expireCookie( 'tree-s' )
		if self.request.cookies.has_key('tree-s'):
		    del self.request.cookies['tree-s']
		del self.request.form['treeCancel']
	    message = ""

	elif self.request.has_key( 'uploadPhoto' ):
	    message = self._uploadPhoto()

        elif self.selectDNKey() != '':
            message = self._selectDN()

        else:
            # You can get here if your form has a single text input
            # field and you press return rather than the form button.
            # Maybe there is a way to figure it out, maybe we don't
            # care.

            message = ""
            if self.request.get('objectFormID','') == 'selectPersonDN':
                message = self._selectDN()

        return message

    # urlencode needs the doseq option that came after Python1.5.2
    # This code comes from Python 2.1.3's urllib.py

    def urlencode(self,query,doseq=0):
        """
        Encode a sequence of two-element tuples or dictionary into a URL 
        query string.
    
        If any values in the query arg are sequences and doseq is true, each
        sequence element is converted to a separate parameter.
    
        If the query arg is a sequence of two-element tuples, the order of the
        parameters in the output will match the order of parameters in the
        input.
        """
    
        if hasattr(query,"items"):
            # mapping objects
            query = query.items()
        else:
            # it's a bother at times that strings and string-like objects are
            # sequences...
            try:
                # non-sequence items should not work with len()
                x = len(query)
                # non-empty strings will fail this
                if len(query) and type(query[0]) != types.TupleType:
                    raise TypeError
                # zero-length sequences of all types will get here and succeed,
                # but that's a minor nit - since the original implementation
                # allowed empty dicts that type of behavior probably should be
                # preserved for consistency
            except TypeError:
                ty,va,tb = sys.exc_info()
                raise TypeError, "not a valid non-string sequence or mapping object", tb
    
        l = []
        if not doseq:
            # preserve old behavior
            for k, v in query:
                k = quote_plus(str(k))
                v = quote_plus(str(v))
                l.append(k + '=' + v)
        else:
            for k, v in query:
                k = quote_plus(str(k))
                if type(v) == types.StringType:
                    v = quote_plus(v)
                    l.append(k + '=' + v)
#                elif type(v) == types.UnicodeType:
#                    # is there a reasonable way to convert to ASCII?
#                    # encode generates a string, but "replace" or "ignore"
#                    # lose information and "strict" can raise UnicodeError
#                    v = quote_plus(v.encode("ASCII","replace"))
#                    l.append(k + '=' + v)
                else:
                    try:
                        # is this a sufficient test for sequence-ness?
                        x = len(v)
                    except TypeError:
                        # not a sequence
                        v = quote_plus(str(v))
                        l.append(k + '=' + v)
                    else:
                        # loop over the sequence
                        for elt in v:
                            l.append(k + '=' + quote_plus(str(elt)))
        return join(l,'&')
    
    def viewJpegPhoto( self ):
	"""
	If this object has a jpegPhoto attribute, set the mime type
        of the response, and return the data.
	"""
        distinguishedName = self.request.get( 'distinguishedName' )
        index = self.request.get( 'index', 0 )
        managedAttributes = self.object._ManagedAttributes
        if managedAttributes.has_key('jpegPhoto'):
            photos = managedAttributes['jpegPhoto'].getValues()
            if photos != []:
		photo = photos[ int(index) ]
		size = len(photo)
		self.request.RESPONSE.setHeader('Content-Type','image/jpeg')
		return photo

	self.request.RESPONSE.setHeader('Content-Type','image/jpeg')
	self.request.RESPONSE.setHeader('Content-Length',0)
	return None
    
    # Internal methods

    def _addObject(self):

        #   Combine items from the REQUEST to form the distinguished
        #   name and object category of the object to be added.  Add
        #   and redirect to a display of it.

        namingAttributeValue = self.request.get( 'namingAttributeValue' )
        parentDN = self.request.get( 'parentDN' )
        if type( parentDN ) == type([]):
            parentDN = parentDN[0]
            self.request.set('parentDN',parentDN)
        parentDN = unquote_plus( parentDN )
        objectCategory = self.request.get( 'objectCategory' )

        distinguishedName = "%s=%s,%s" % ( self.namingAttribute,
                                           namingAttributeValue,
                                           parentDN )

        try:
            self.object.create( distinguishedName, objectCategory )
            message = "Added object '%s'.  " % ( distinguishedName )
	    if self.request.has_key('unitTest'):
                return message
            else:
                url = "%s?distinguishedName=%s&manage_tabs_message=%s" % \
                      ( self.request['URL0'], 
                        quote_plus(distinguishedName), 
                        quote_plus(message) )
                self.request.RESPONSE.redirect( url )

        except ALREADY_EXISTS:
            message = """Object '<a href="?distinguishedName=%s">%s</a>' 
                already exists!""" % ( distinguishedName, distinguishedName )
            return message

        # any other exception will raise a Zope error

    def _delObject(self):
        if not self.request.has_key('areYouSure'):
            self.request.set( 'areYouSure', '1' )
            return "ARE YOU SURE?"

        elif self.request.has_key('ids'):
            ids = self.request.get( 'ids', [] )
            for distinguishedName in ids:
#                self.object.refresh( distinguishedName )
                self.object.delete( self.request.get('areYouSure'),
                                    distinguishedName = distinguishedName )
        else:
            targetDN = self.request.get('targetDN') 
            self.object.refresh( targetDN )
            self.object.delete( self.request.get('areYouSure') )

	del self.request.form['updateObject']
	del self.request.form['ids']
	del self.request.form['areYouSure']
        return "Objects deleted."

    def _renameObject(self):

        #   Rename the object specified by objectDN in the request,
        #   and rename it with the new namingAttributeValue

        objectDN = unquote_plus( self.request.get( 'objectDN' )[0] )
        objectCategory = self.request.get( 'objectCategory' )
	namingAttribute = split( objectDN, '=' )[0]
        newRDN = "%s=%s" % (namingAttribute,
                            self.request.get( 'namingAttributeValue' )
                            )
        newDN = "%s,%s" % ( newRDN,
                            getParentFromDN( objectDN ) )

        try:
            self.user.rename( objectDN, newRDN )
            message = "Renamed object '%s'.  " % ( objectDN )
	    if self.request.has_key('unitTest'):
                return message
            else:
                url = "%s?distinguishedName=%s&manage_tabs_message=%s" % \
                      ( self.request['URL0'], 
                        quote_plus(newDN), 
                        quote_plus(message) )
                self.request.RESPONSE.redirect( url )

        except ALREADY_EXISTS:
            message = """Object '<a href="?distinguishedName=%s">%s</a>' already exists!""" % ( newDN, newDN )
            return message

        # any other exception will raise a Zope error

    def _renamePerson(self):

        #   If we got here from the _peopleRenameForm, then
        #   perform the rename.

        if self.request.has_key('haveNewNames'):
            currentDNs = self.request.get( 'ids', [] )
            newNames = self.request.get( 'new_ids', [] )

            # Skip empty new_ids

            while '' in newNames:
                i = index(newNames,'')
                currentDNs.pop(i)
                newNames.pop(i)

            message = "Renamed "
            for i in range( 0, len( currentDNs ) ):
                currentDN = currentDNs[ i ]
                newName = newNames[ i ]

                newRDN = "%s=%s" % (self.namingAttribute, newName )
                objectCategory = self.request.get( 'objectCategory' )
                newDN = "%s,%s" % ( newRDN,
                                    getParentFromDN( currentDN ) )

                try:
                    self.user.rename( currentDN, newRDN )
                    message = message + "'%s' to %s " % ( currentDN, newDN )

                except ALREADY_EXISTS:
                    message = message + """Object '<a href="?distinguishedName=%s">%s</a>' 
                    already exists!""" % ( newDN, newDN )


            self.request.objectFormID = "renameObject"
            del self.request.form['updateObject']
            return message

        #   Otherwise, set REQUEST variables so that the
        #   _renamePersonForm gets displayed.

        else:
            self.request.set('objectFormID', 'renamePerson')
            del self.request.form['updateObject']

            return ""




    def _addMembers(self):
        ids = self.request.get( 'ids', [] )
        # Am I to add a 'member' or 'uniqueMember'?

        objectClassAttr = self.object.getManagedAttribute('objectClass')
        if 'groupOfNames' in objectClassAttr.getRequiredValues():
            memberAttrName = 'member'
        elif 'groupOfUniqueNames' in objectClassAttr.getRequiredValues():
            memberAttrName = 'uniqueMember'
        else:
            memberAttrName = None
        if memberAttrName:
            memberAttr = self.object.getManagedAttribute(memberAttrName)

        for distinguishedName in ids:
            memberAttr.addValue( distinguishedName )

        if len( ids ) > 0:
            self.object.updateDirectory()
            message = "Added group members %s" % ids
        else:
            message = ""

        self.request.set( 'ids', [] )
        return message

    def _removeMembers(self):
        if not self.request.get('areYouSure',0):
            self.request.set( 'areYouSure', 1 )
            return "ARE YOU SURE?"

        elif self.request.has_key('ids'):
            self.request.set('areYouSure',0)
            objectClassAttr = self.object.getManagedAttribute('objectClass')
            if 'groupOfNames' in objectClassAttr.getRequiredValues():
                memberAttr = self.object.getManagedAttribute('member')
            elif 'groupOfUniqueNames' in objectClassAttr.getRequiredValues():
                memberAttr = self.object.getManagedAttribute('uniqueMember')
            else:
                memberAttr = None
            ids = self.request.get( 'ids', [] )
            if memberAttr is not None:
                for distinguishedName in ids:
                    memberAttr.delValue( distinguishedName )

            if len( ids ) > 0:
                self.object.updateDirectory()
                message = "Removed group members"
            else:
                message = ""

        self.request.set( 'ids', [] )
        return message

    def _selectDN(self):

        targetAttribute = self.request.get('targetAttribute',self.selectDNKey())
        option = self.request.get( 'selectDN_' + targetAttribute )
        objectCategory = self.request.get( 'objectCategory' )

        #   If we got here from the _selectDNForm then update the REQUEST.
        #   Furthermore, refresh my object with the selected DN, if
        #   we were selecting an OrgUnit or group, which are selected
        #   singly.

        if option in [" Nobody "," None "]:
            self.request.set( targetAttribute, '' )

            if self.request.form.has_key('targetAttribute'):
                del self.request.form['targetAttribute']
            if self.request.has_key('returnsToFormID'):
                self.request.set( 'objectFormID', 
                                  self.request.get('returnsToFormID') )
            else:
                self.request.set( 'objectFormID', 'display' )
            return "Selected %s.  You still need to Save." % ( targetAttribute )

        #   "Select" is a tricky value.  If a Select button is taking
        #   us away from objectFormID to a screen for selecting, then
        #   just go on.  But if a Select button is taking us back to
        #   objectFormID, then process it here.

        objectFormID = self.request.get('objectFormID')
        returnsToFormID = self.request.get('returnsToFormID')
        if option == " Select " and objectFormID != returnsToFormID:
            self.request.set( targetAttribute, self.request.get( 'ids' ) )

            if self.request.form.has_key('targetAttribute'):
                del self.request.form['targetAttribute']
            self.request.set( 'objectFormID', 'display' )
            return "Selected %s.  You still need to Save." % ( targetAttribute )

        else:
            self.request.set( 'targetAttribute', targetAttribute )

            if self.request.has_key( 'pointsTo' ):
                pointsTo = self.request['pointsTo']

            elif hasattr( self, 'schema' ) and \
             targetAttribute in self.schema.getManagedAttributeNames():
                attrSchema = self.schema.getAttributeSchema(targetAttribute)
                pointsTo = attrSchema.getPointsTo()

            else:
                pointsTo = 'orgUnit'

            #   Set REQUEST variables so that one of the
            #   _select(objectCategory)DNForm gets
            #   displayed next.

            if lower(pointsTo) == 'person':
                self.request.set('objectFormID', 'selectPersonDN' )
            elif lower(pointsTo) == 'personorgunit':
                self.request.set('objectFormID', 'selectPersonOrgUnitDN')
            elif lower(pointsTo) == 'orgunit':
                self.request.set('objectFormID', 'selectOrgUnitDN' )
            elif lower(pointsTo) == 'group':
                self.request.set('objectFormID', 'selectGroupDN' )
            return self.request['objectFormID']

    def _selectObjectByBrowsing(self):

        #   Unless the Cancel button was pressed, set REQUEST
        #   variable for showing the selectPersonByBrowsing form.

        if self.request.has_key('treeCancel'):

            if self.request.form.has_key('query') and \
             self.request.form['query'] == '':
                del self.request.form['query']
            if self.request.other.has_key('query') and \
             self.request.other['query'] == '':
                del self.request.other['query']

        else:

            returnsToFormID = self.request.get( 'returnsToFormID',
                                          self.request.get('objectFormID'))

            browseObjectCategory = \
		self.request.get('browseObjectCategory','person')

            if browseObjectCategory == 'person':
                self.request.set('objectFormID','selectPersonByBrowsing')
            elif browseObjectCategory == 'orgUnit':
                self.request.set('objectFormID', 'selectOrgUnitByBrowsing')
            elif browseObjectCategory == 'group':
                self.request.set('objectFormID', 'selectGroupByBrowsing')
            self.request.set('returnsToFormID', returnsToFormID )
            return self.request['objectFormID']

    def _uploadPhoto(self):

	# Update the jpegPhoto attribute, possibly by uploading
	# a file.  The normal method of putting values into the
	# request is not used for this attribute.

	# Get values from form

        option = self.request.get( 'uploadPhoto' )
        distinguishedName = self.request.get( 'distinguishedName' )
        index = int(self.request.get( 'index', 0 ))
        displayCategory = self.request.get( 'displayCategory' )
        objectFormID = self.request.get('objectFormID')
        returnsToFormID = self.request.get('returnsToFormID')

        if self.request.has_key('file') and self.request['file']:
            photo = self.request.get('file').read()
            size = len(photo)
	else:
	    photo = None
	    size = 0

	# We impose a limit on the total byte count of all the
	# photos (unless maxJpegPhotoSize is 0).  See how close we are.

        jpegObject = self.object.getManagedAttribute( 'jpegPhoto' )
	oldPhotos = jpegObject.getValues()
	totalSize = 0
	for item in oldPhotos:
	    totalSize = totalSize + len(item)

        if size+totalSize > self.maxJpegPhotoSize and self.maxJpegPhotoSize:
            self.request.set('manage_tabs_message',
		"There are already %s bytes of photos here.  " % totalSize +
		"This photo's size (%s bytes) would put the total over the limit of  %s bytes." % \
                       ( size, self.maxJpegPhotoSize ) +
                " Please pick a photo smaller than %s bytes." % \
                       ( self.maxJpegPhotoSize-totalSize ) )
            return self.request['objectFormID']
        elif size > self.maxJpegPhotoSize and self.maxJpegPhotoSize:
            self.request.set('manage_tabs_message',
		"""Photo size (%s bytes) is too big.  Please pick a photo smaller than %s bytes.""" % \
                       ( size, self.maxJpegPhotoSize ) )
            return self.request['objectFormID']

	# Apply the selected option

        self.request.set('objectFormID',returnsToFormID)

	if photo and ( option == ' Upload/Add ' or \
	 ( option==' Upload/Replace ' and len(oldPhotos)==0 ) ):
            jpegObject.addValue( photo )
            self.object.setManagedAttribute( jpegObject )
            self.object.updateDirectory()
            return "Updated photo.  You still need to Save any other " + \
                   "changed attributes."

	elif photo and option == ' Upload/Replace ':
	    oldPhoto = jpegObject.getValues()[index]
            jpegObject.delValue( oldPhoto )
            jpegObject.addValue( photo )
            self.object.setManagedAttribute( jpegObject )
            self.object.updateDirectory()
            return "Updated photo.  You still need to Save any other " + \
                "changed attributes."

	elif option == ' Delete ' and returnsToFormID != objectFormID:
	    oldPhoto = jpegObject.getValues()[index]
            jpegObject.delValue( oldPhoto )
            self.object.setManagedAttribute( jpegObject )
            self.object.updateDirectory()
            self.request.set('objectFormID',returnsToFormID)
            return "Deleted photo.  You still need to Save any other " + \
		"changed attributes."

        self.request.set('objectFormID','updateJpegPhoto')
        return ""

    def _updateObjectValues(self):

        #   Pull out all of the ManagedAttributes in this displayCategory
        #   that the user can write.  Update them with values from the
        #   REQUEST, put them back, and update the directory.

        distinguishedName = self.request.get( 'distinguishedName' )
        displayCategory = self.request.get( 'displayCategory' )
        object = self.object
        object.refresh()

        attrNames = object.getManagedAttributeNames( displayCategory )
        attrNames = filter( lambda i,o=object: o.canWriteAttribute(i),
                            attrNames )

	# jpegPhoto is an exception, updated on it's own form.

	if 'jpegPhoto' in attrNames:
	    attrNames.remove( 'jpegPhoto')

        for attrName in attrNames:
            attrObject = object.getManagedAttribute( attrName )

            #   Special case for paswords, which need two values in
            #   the REQUEST.

            if attrObject.getSyntaxName() == 'Password':
                pw1 = self.request.get( attrName+'1', '' )
                pw2 = self.request.get( attrName+'2', '' )

                if pw1 != pw2:
                    self.request.set( attrName+'1', 'Password' )
                    self.request.set( attrName+'2', 'Password' )
                    return "Passwords do not match.  Try again."

                elif pw1 == 'Password':
                    break

                else:
                    attrObject.setValues( [pw1,] )

                    # Did I change my own password?

                    myDN = getSecurityManager().getUser().getUserDN()
                    if DNsAreEquivalent( myDN, distinguishedName ):
                        self._IChangedMyOwnPassword = TRUE

            else:
                attrObject.setValues( self.request.get( attrName, [] ) )

            object.setManagedAttribute( attrObject )

        object.updateDirectory()
        return "Updated values in the directory."

