"""

    Module: $RCSfile: ZReports.py,v $
    Author: $Author: payne $
    Date:   $Date: 2003/12/15 00:47:05 $

    The ZReports class is the mutable schema object class.  It holds
    an AuthenticatedUser instance, a ManagedSchema class instance,
    a ManagedRoles instance, the REQUEST, and contains methods used 
    on the various faces of the Reports tab.

    The Reports tab provides a number of reports, derived from the
    configuration of the LDAP Directory Manager:

      o Schema Documentation in HTML format, suitable for printing
	or viewing on the web.  All of the configured roles are listed
	along with the roles associated with each permission.  Then
	for each of Person, Org Unit, and Group schema, the roles
	granted access to the Add/Delete object links, are listed,
	followed by an list of all of the managed attributes and
	their properties.

      o	Security assertions in OpenLDAP slapd.conf format, suitable
	for applying to the directory to enforce/enable the access
	and permissions specified by the configuration of the LDAP
	Directory Manager.  This makes configuring the directory to 
	match the directory manager trivial, whereas the reverse is not.

      o Security assertions in Netscape/iPlanet/SunONE ACIs in LDIF 
	format, suitable for importing into the directory to enforce/
	enable the access and permissions specified by the configuration 
	of the LDAP Directory Manager.  This makes configuring the
	directory to match the directory manager trivial, whereas
	the reverse is not.  Provision is made for recording ACIs
	not related to the configuration of the directory manager
	that need to be applied to the same entry.

      o Configuration in the format of the SchemaDefaults.py file.
	By replacing the file in the distribution with this output
	from a configured LDAP Directory Manager insures that any
	subsequent instantiations of the product will have the same
	configuration.  This can be especially helpful in updating
	to more recent versions of the product.

Copyright (C) 2002-2003 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

"""
__version__ = '$Revision: 1.8 $'[11:-2]

from AccessControl import ClassSecurityInfo
from AccessControl.Permission import Permission
from AccessControl.Role import RoleManager
from Acquisition import Implicit
from copy import deepcopy
from Globals import DTMLFile
from string import join, lower

from LDAPExceptions import REQUIRED_ATTRIBUTE, NO_SUCH_ATTRIBUTE
from LDAPUtilities import _ldap, DNsAreEquivalent, \
    list1ContainsList2, mergeLists, TRUE, FALSE
from LDIFObjectParser import LDIFObjectParser
from LDIFAttributeParser import LDIFAttributeParser
from ManagedAttributeSchema import ManagedAttributeSchema
from ZRoles import ZRoles

class ZReports( Implicit ):

    security = ClassSecurityInfo()
    security.declareProtected('LDAP Manage Schema', 
        'manage_schemaFormSelection',
        )

    manage_reportFormSelection = DTMLFile('dtml/reportFormSelection', globals())

    #   DTML templates for the various "faces" of the Schema tab.
    #   Each template's forms must contain a hidden variable that
    #   identifies the form.  The name is 'reportsFormID', and the
    #   value must match one of the following after prepending an
    #   underscore, and appending 'Form', as in:
    #
    #     <input type=hidden name=reportsFormID value=schemaAttributes>
 
    getSchemaReport = DTMLFile('dtml/schemaReport', globals())
    getSchemaAttrReport = DTMLFile('dtml/schemaAttrReport', globals())
    _reportOpenLDAPForm = DTMLFile('dtml/reportOpenLDAP',globals())
    _reportIPlanetForm = DTMLFile('dtml/reportIPlanet',globals())
    _reportSchemaDefaultsForm = DTMLFile('dtml/reportSchemaDefaults',
                                          globals())
    _reportSchemaDocForm = DTMLFile('dtml/reportSchemaDoc', globals())

    def __init__(self, user, REQUEST ):

        #   Update the Zope schema object with values from the LDAP 
        #   schema (and save the list of available LDAP object classes).  
        
        self.user = user
        self.request = REQUEST

        if not self.request.has_key('objectCategory'):
            self.request.set('objectCategory','person')
        category = lower(self.request['objectCategory'])

        if category == 'group':
            self._schema = user.getLDAPDirectory().getGroupSchema()
        elif category == 'person':
            self._schema = user.getLDAPDirectory().getPersonSchema()
        elif category == 'orgunit':
            self._schema = user.getLDAPDirectory().getOrgUnitSchema()
        else:
            assert FALSE            

        self._managedRoles = user.getLDAPDirectory().getManagedRoles()

    def DNsAreEquivalent( self, dn1, dn2):
        """
           Wrapper for LDAPUtilities.DNsAreEquivalent
        """
        return DNsAreEquivalent( dn1, dn2 )

    def getManagedRoles(self):
        """
            Returns a managed roles object.
        """
        return self._managedRoles

    def getManagedSchema(self):
        """
        Returns a managed schema object,
        """
        return self._schema

    def getOpenLDAPAccessByKey( self, roles, dn, attrNameList, searchFilter,
				key ):
	#
	# Utility method for generating OpenLDAP access rule output
	# given a key encoding the permissions, and a label set to
	# 'read' or 'write'
	#

        zopeRoleSearchBase = \
            self.user.getLDAPDirectory().getZopeRoleSearchBase()
        writeKey = key>>len(roles)
        readKey = key - (writeKey<<len(roles))

        output = """access to dn.subtree="%s"
          filter=%s
          attrs=%s""" % ( dn, searchFilter, join(attrNameList,',') )

        for role in roles:
            mask = 1<<(len(roles) - roles.index(role) - 1 )
            flag = mask & writeKey
            if flag:
                if role == 'SELF':
                    output = output + """
          by self write"""
                elif role == 'MNGR':
                    output = output + """
          by dnattr=manager write"""
                elif role == 'IFALLOWED':
                    pass
                elif role not in ['AUTH','ANON']:
                    for group in self.user.getMatchingGroups( 
                      role, zopeRoleSearchBase ):
                        if self.user.whatKindOfGroup(group) == \
                          'groupOfUniqueNames':
                            output = output + """
          by group/groupOfUniqueNames/uniqueMember="%s" write""" % group
                        else:
                            output = output + """
          by group="%s" write""" %  group
                elif role == 'AUTH':
                    output = output + """
          by users write"""
                elif role == 'ANON':
                    output = output + """
          by anonymous write""" 

        for role in roles:
            mask = 1<<(len(roles) - roles.index(role) - 1 )
            writeFlag = mask & writeKey
            readFlag = mask & readKey
            if readFlag and not writeFlag:
                if role == 'SELF':
                    output = output + """
          by self read"""
                elif role == 'MNGR':
                    output = output + """
          by dnattr=manager read"""
                elif role == 'IFALLOWED':
                    pass
                elif role not in ['AUTH','ANON']:
                    for group in self.user.getMatchingGroups( 
                          role, zopeRoleSearchBase ):
                        if self.user.whatKindOfGroup(group) == \
                              'groupOfUniqueNames':
                                output = output + """
          by group/groupOfUniqueNames/uniqueMember="%s" read""" % group
                        else:
                            output = output + """
          by group="%s" read""" % group
                elif role == 'AUTH':
                    output = output + """
          by users read"""
                elif role == 'ANON':
                    output = output + """
          by anonymous read"""
	return output

    def getSchemaPermission(self):
        """
        Returns the name of the permission for managing the schema
        """
        return self.manageSchemaPermission

    def getSecurityAttrConf( self, roles, schema, dn, searchFilter ):
        """
        Given a schema and a list roles, create OpenLDAP slapd.conf
        lines describing the access controls.
        """
        # Build a dictionary keyed by access rules, to sort all of the
        # attributes with the same access rules into the same security
        # assertion.

        dict = self.getSecurityAttrDict( roles, schema )

        # For each key, build a complete OpenLDAP security assertion,
        # converting the key into assertion text.

        output = ''
	namingAttribute = schema.getNamingAttribute()

        for key in dict.keys():
            attrNameList = dict[key]

	    # The namingAttribute is controlled in getSecurityControllerConf

	    if namingAttribute in attrNameList:
		attrNameList.remove(namingAttribute)

	    if attrNameList and 'userPassword' in attrNameList:

		attrNameList.remove('userPassword')
		if attrNameList:
		    output = output + \
    		        self.getOpenLDAPAccessByKey( roles, dn, attrNameList, 
						     searchFilter, key )
		    output = output + '\n'

		output = output + \
    		    self.getOpenLDAPAccessByKey( roles, dn, ['userPassword'], 
						 searchFilter, key )
        	writeKey = key>>len(roles)
        	for role in roles:
		    mask = 1<<(len(roles) - roles.index(role) - 1 )
		    writeFlag = mask & writeKey
        	    writeKey = key>>len(roles)
                    if not writeFlag:
                        if role == 'SELF':
                            output = output + """
          by self auth""" 
                        elif role == 'AUTH':
                            output = output + """
          by users auth"""
                        elif role == 'ANON':
                            output = output + """
          by anonymous auth"""
                output = output+'\n\n'

	    elif attrNameList:

		output = output + \
    		    self.getOpenLDAPAccessByKey( roles, dn, attrNameList, 
						 searchFilter, key )

                output = output+'\n\n'

        return output

    def getSecurityAttrDict( self, roles, schema ):
        """
        Build a dictionary whose key encodes read/write access by all
        of the permanent roles and the roles that map to LDAP groups,
        and whose value is a list of attribute ldapNames that have the
        encoded permissions.
        """
        dict = {}
	anonymousRoleName = \
	    self._managedRoles.getRoleNameFromCanonicalName( 'ANON' )
        for attrName in schema.getManagedAttributeNames():
            attr = schema.getAttributeSchema( attrName )
            writers = deepcopy(attr.getWriters())
            zopeWriters = self.zopeRoleNeedsANON( anonymousRoleName, writers )
	    if zopeWriters:
		writers.append('ANON')
            readers = deepcopy(attr.getReaders())
	    zopeReaders = self.zopeRoleNeedsANON( anonymousRoleName, readers )
	    if zopeReaders:
		readers.append('ANON')
            writeKey = 0
            readKey = 0
            for role in roles:
                if role in writers:
                    writeKey = (writeKey<<1) + 1
                    readKey = (readKey<<1)
                elif role in readers:
                    writeKey = (writeKey<<1)
                    readKey = (readKey<<1) + 1
                else:
                    writeKey = (writeKey<<1)
                    readKey = (readKey<<1)

            key = (writeKey<<len(roles)) + readKey

            if dict.has_key( key ):
                dict[key].append( attrName )
            else:
                dict[key] = [attrName]
        return dict

    def getSecurityControllerConf(self, schema, dn, searchFilter, groupBase ):
        """
        Write OpenLDAP slapd.conf lines describing what the schema
        controllers can do.
        """
	namingAttribute = schema.getNamingAttribute()
        output = """access to dn.subtree="%s"
          filter=%s
          attrs=children,entry,%s""" % ( dn, searchFilter, namingAttribute )

        controllers = []
        roles = self.rolesOfPermission( schema.getPermissionName() )
        for role in roles:
            if role['selected'] == 'SELECTED':
                controllers.append( role['name'] )
        for controller in controllers:
            for controllerDN in self.user.getMatchingGroups( controller, groupBase ):
                if self.user.whatKindOfGroup(controllerDN) == 'groupOfUniqueNames':
                    output = output + """
          by group/groupOfUniqueNames/uniqueMember="%s" write""" % controllerDN
                else:
                    output = output + """
          by group="%s" write""" % controllerDN

        output = output + """
          by * read\n\n"""
        return output

    def getSecurityOpenLDAPConf(self):
        """
        Return the configured security (access) settings in a format
        suitable for use in OpenLDAP's slapd.conf file.
        """
        # Make sure everyone can read the schema

        output = """# Start of LDAP Manager generated configuration

access to dn="" by * read

access to dn="cn=subschema" by * read\n\n"""

        # We need to arrange the available roles in order of most
        # specific to least specific.  We cheat here, and put the
        # permanent roles in order, assuming that we know what they
        # are, rather than taking them from the configuration -- BAD.
        # We put the configured roles just ahead of 'AUTH', and
        # ignore 'IFALLOWED'.

        roles = ['SELF','MNGR',] + \
                self._managedRoles.getConfiguredRoleNames() + \
                ['AUTH','ANON']

        # By object category, we make use of a dictionary whose key 
        # encodes read/write access by all of the permanent roles and 
        # the roles that map to LDAP groups, and whose value is a list
        # of attribute ldapNames that have the encoded permissions.

        groupBase = self.user.LDAPDirectory.getGroupBrowseBase()

        # 'person'
        schema = self.user.LDAPDirectory._personManagedSchema
        dn = self.user.LDAPDirectory.getDirectoryBase()
        searchFilter = '(objectclass=person)'
        output = output + self.getSecurityControllerConf( schema, dn, searchFilter,
                                                          groupBase )
        output = output + self.getSecurityAttrConf( roles, schema, dn, searchFilter )

        # 'orgUnit'
        schema = self.user.LDAPDirectory._orgUnitManagedSchema
        dn = self.user.LDAPDirectory.getDirectoryBase()
        searchFilter = '(objectclass=organizationalunit)'
        output = output + self.getSecurityControllerConf( schema, dn, searchFilter,
                                                          groupBase )
        output = output + self.getSecurityAttrConf( roles, schema, dn, searchFilter )

        # 'group'
        schema = self.user.LDAPDirectory._groupManagedSchema
        dn = self.user.LDAPDirectory.getDirectoryBase()
        searchFilter = '(|(objectclass=groupofnames)(objectclass=groupofuniquenames))'
        output = output + self.getSecurityControllerConf( schema, dn, searchFilter,
                                                          groupBase )
        output = output + self.getSecurityAttrConf( roles, schema, dn, searchFilter )

        output = output + """
# End of LDAP Manager generated configuration"""
        return output

    def getSecurityIPlanetLDIF(self):
        """
            Returns the configured security (access) settings in LDIF
            format suitable for use iPlanet Directory Server aci's.
        """

        # First, the readable/writeable attributes for the permanent 
        # roles and the roles that map to LDAP groups, and then the
        # controllers, by objectCategory...  Note that the 'IFALLOWED' 
        # role is not implemented.
        # 'person'

        schema = self.user.LDAPDirectory._personManagedSchema
        dn = self.user.LDAPDirectory.getDirectoryBase()
        groupdn = self.user.LDAPDirectory.getGroupBrowseBase()
        searchFilter = '(objectclass=person)'
        objectCategory = 'person'
        output = """dn: %s
changetype: modify
delete: aci
-"""        % ( dn )
        aciLines = 0
        for aciLine in self.user.LDAPDirectory.getIPlanetACIs():
            if aciLine:
                output = output + "\n%s" % aciLine
                aciLines = aciLines + 1
        if aciLines:
            output = output + '\n-'
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'SELF', 'userdn', 'ldap:///self' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'ANON', 'userdn', 'ldap:///anyone' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'AUTH', 'userdn', 'ldap:///all' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'MNGR', 'userattr', 'manager#USERDN' )

        for role in self._managedRoles.getConfiguredRoleNames():
            for roleDN in self.user.getMatchingGroups( role, groupdn ):
                output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                                    objectCategory,
                                                    role, 'groupdn',
                                                    'ldap:///'+roleDN )
        output = self.iPlanetControllerLDIF( output, schema, dn, groupdn, 
                                             searchFilter, objectCategory )

        # 'group'

        schema = self.user.LDAPDirectory._groupManagedSchema
        dn = self.user.LDAPDirectory.getDirectoryBase()
        searchFilter = '(|(objectclass=groupofnames)(objectclass=groupofuniquenames))'
        objectCategory = 'group'
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'SELF', 'userdn', 'ldap:///self' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'ANON', 'userdn', 'ldap:///anyone' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'AUTH', 'userdn', 'ldap:///all' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'MNGR', 'userattr', 'manager#USERDN' )
        for role in self._managedRoles.getConfiguredRoleNames():
            for roleDN in self.user.getMatchingGroups( role, groupdn ):
                output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                                    objectCategory,
                                                    role, 'groupdn',
                                                    'ldap:///'+roleDN )
        output = self.iPlanetControllerLDIF( output, schema, dn, groupdn, 
                                             searchFilter, objectCategory )

        # 'orgUnit'

        schema = self.user.LDAPDirectory._orgUnitManagedSchema
        dn = self.user.LDAPDirectory.getDirectoryBase()
        searchFilter = '(objectclass=organizationalUnit)'
        objectCategory = 'orgUnit'
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'SELF', 'userdn', 'ldap:///self' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'ANON', 'userdn', 'ldap:///anyone' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'AUTH', 'userdn', 'ldap:///all' )
        output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                            objectCategory,
                                            'MNGR', 'userattr', 'manager#USERDN' )
        for role in self._managedRoles.getConfiguredRoleNames():
            for roleDN in self.user.getMatchingGroups( role, groupdn ):
                output = self.iPlanetAttributeLDIF( output, schema, dn, searchFilter,
                                                    objectCategory,
                                                    role, 'groupdn',
                                                    'ldap:///'+roleDN )
        output = self.iPlanetControllerLDIF( output, schema, dn, groupdn, 
                                             searchFilter, objectCategory )
        return output

    def getSchemaForSchemaDefaults( self, schema ):
        """
        Write out attribute schemas for the supplied schema in
        SchemaDefaults format.

        """
        output = ""
        for ldapName in schema.getManagedAttributeNames():
            ldapAttr = schema.getAttributeSchema( ldapName )
            output = output + "\n        '%s' : {" % ldapName
            if ldapAttr.getFriendlyName() != ldapName:
                output = output + """
                'friendlyName' : '%s',""" % ldapAttr.getFriendlyName()
            if ldapAttr.getDescription():
                output = output + """
                'description' : \"\"\"%s\"\"\",""" % ldapAttr.getDescription()
            output = output + """
                'readers' : %s,
                'writers' : %s,
                'displayCategories' : %s,""" \
                % ( ldapAttr.getReaders(),
                    ldapAttr.getWriters(),
                    ldapAttr.getDisplayCategories(), )
            if ldapAttr.hasValidValues():
                output = output + """
                'validValues' : %s,""" % ldapAttr.getValidValues()
            if ldapAttr.getRequiredValues() != []:
                output = output + """
                'requiredValues' : %s,""" % ldapAttr.getRequiredValues()
            if ldapAttr.getPointsTo() != '':
                output = output + """
                'pointsTo' : '%s',""" % ldapAttr.getPointsTo()
            if ldapAttr.getRequired():
                output = output + """
                'required' : %s,""" % ldapAttr.getRequired()
            output = output + " },\n"
        output = output + "        }\n"
        return output

    def getSecuritySchemaDefaults(self):
        """
        Write out the security assertions (actually all of the 
        configuration) in a format suitable for replacing the
        SchemaDefaults module.

        """
        directory = self.user.LDAPDirectory
        personSchema = directory.getPersonSchema()
        groupSchema = directory.getGroupSchema()
        orgUnitSchema = directory.getOrgUnitSchema()
        managedRoles = directory.getManagedRoles()

        output = """\"\"\" SchemaDefaults.py, automatically generated. 

    The STLDAPManager is designed to be highly configurable, but
configuring an instance can be a lot of work, which you will not
want to repeat.  This module, WHICH IS ONLY USED WHEN CREATING A NEW
LDAP Directory Manager, does provide a way to permanently record a
configuration.  The sensible way to configure a directory manager is
through-the-web.  When you are finished, you can save your configuration
in SchemaDefaults format, replacing this module.

    However, there are a few items that you may wish to configure the
first time, and they appear near the the top of the module.  Changing
things here can cause unit tests to fail, however.
\"\"\"

#   LDAP Server
#
#   Basic properties of the LDAP directory:
#
#     1. The server hostname and port (the port need not be specified
#        if it is the default 389).
#
#     2. The DN at the root of the tree to be managed by this instance.
#
#     3. The DN at the root of the tree to be searched when looking for
#        people.
#
#     4. The DN at the root of a tree containing organizational units
#        arranged hierarchically to match your org chart.  Children of
#        this DN should have 'organizationalUnit' as an objectClass 
#        value.
#
#     5. The DN at the root of a tree of groups.  Children of this DN
#        should have 'groupOfNames' or 'groupOfUniqueNames' as a value
#        of objectClass.
#
#     6. The DN at the root of a tree of groups used for determining
#        Zope roles.  Children of this DN should have 'groupOfNames'
#        or 'groupOfUniqueNames' as an objectClass value.
#
#     7. The scope of a search, starting at #6, for finding Zope roles
#        where:
#                 0 <=> 'Search the base ONLY'
#                 1 <=> 'Search one level of children'
#                 2 <=> 'Search the base and all subtrees'
#
#
#   Items #3 to #6 default to the value of #2.
"""

        output = output + """
defaultLDAPServer = %s
defaultDirectoryBase = '%s'
defaultPersonSearchBase = '%s'
defaultOrgUnitBrowseBase = '%s'
defaultGroupBrowseBase = '%s'
defaultZopeRoleSearchBase = '%s'
defaultZopeRoleSearchScope = %s
""" % \
        ( directory.getServers(),
          directory.getDirectoryBase(),
          directory.getPersonSearchBase(),
          directory.getOrgUnitBrowseBase(),
          directory.getGroupBrowseBase(),
          directory.getZopeRoleSearchBase(),
          directory.getZopeRoleSearchScope(), )

	output = output + """
#  If you are using the SunONE/iPlanet/Netscape directory server
#   enter the ACIs on the directory root object.  This will allow
#   the directory manager to generate an LDIF file that sets all
#   of the ACIs for the directory.  (Modifying the ACI attribute
#   doesn't work, because the directory manager does not know
#   what the ACI to be replaced looks like.  So you have to 
#   delete the attribute and add ACIs back in.)  Preceed each
#   ACI with the line (don't indent):
#
#   add: aci
#
#   and separate ACIs with a line consisting only of a hyphen, as
#   in (don't indent):
#
#    add:aci
#    aci:
#     (target = "ldap:///o=test")
#     (targetattr = "*")
#     (version 3.0;
#     acl "Configuration Administrator";
#     allow (all)
#     (userdn = "ldap:///uid=admin,ou=Administrators, ou=TopologyManagement,
#     o=NetscapeRoot");)
#    -
#    add:aci
#    aci:
#     (target = "ldap:///o=test")
#     (targetattr = "*")
#     (version 3.0;
#     acl "Configuration Administrators Group";
#     allow (all)
#     (groupdn = "ldap:///cn=Configuration Administrators, ou=Groups, 
#     ou=TopologyManagement,o=NetscapeRoot");)

defaultIPlanetACIs = \"\"\""""

        for aciLine in directory.getIPlanetACIs():
            if aciLine:
                output = output + "%s\n" % aciLine
        output = output + '"""\n'

        groupObjectClass = ''
        for objectClass in \
         groupSchema.getAttributeSchema('objectClass').getRequiredValues():
            if objectClass[:5] == 'group':
                groupObjectClass = objectClass
                break
          
        output = output + """
#   Configured roles map to group names or acquired Zope role names, and
#   canonical roles are permanent roles that map to categories of
#   directory access: 

defaultConfiguredRoleNames = %s
defaultCanonicalRoleNames = %s
""" %   ( managedRoles.getConfiguredRoleNames(),
          managedRoles.getPermanentRoles(),)

        output = output + """
#   personsOUAttr is the key for determing a person's org unit: 'dn' if
#   you look at a person's dn, or the name of the person attribute that
#   contains the person's org unit.
#   groupObjectClass can be either 'groupOfNames' or 'groupOfUniqueNames'.

defaultPersonsOUAttr = '%s'
defaultGroupObjectClass = '%s'

#   Controllers are the roles that can add/delete entries of a given
#   object category.

defaultGroupControllers = %s
defaultOrgUnitControllers = %s
defaultUserControllers = %s
""" %   ( personSchema.getPersonsOUAttr(),
          groupObjectClass,
          groupSchema.getControllers(),
          orgUnitSchema.getControllers(),
          personSchema.getControllers(), )

	peopleRoleSelected = filter(lambda i: i['selected']=='SELECTED',
	    self.aq_parent.rolesOfPermission(self.managePeoplePermission))
	groupRoleSelected = filter(lambda i: i['selected']=='SELECTED',
	    self.aq_parent.rolesOfPermission(self.manageGroupPermission))
	orgUnitRoleSelected = filter(lambda i: i['selected']=='SELECTED',
	    self.aq_parent.rolesOfPermission(self.manageOrgUnitPermission))
	schemaRoleSelected = filter(lambda i: i['selected']=='SELECTED',
	    self.aq_parent.rolesOfPermission(self.manageSchemaPermission))
        output = output + """
#   Roles that are not controllers may nevertheless have permission to
#   view tabs.  These are the roles that have access to the tabs.

defaultManagePeopleRoles = %s
defaultManageGroupRoles = %s
defaultManageOrgUnitRoles = %s
defaultManageSchemaRoles = %s
""" %   ( map( lambda i: i['name'], peopleRoleSelected ),
          map( lambda i: i['name'], groupRoleSelected ),
          map( lambda i: i['name'], orgUnitRoleSelected ),
          map( lambda i: i['name'], schemaRoleSelected ),)

        output = output + """
#   The permanentRoles dictionary contains a configurable roleName
#   and a description for each role.

defaultPermanentRoles = {
"""
        for role in managedRoles.getPermanentRoles():
            output = output + """
        '%s' : {
                'roleName'     : '%s',
                'description'  : \"\"\"%s\"\"\" },
""" %       ( role,
              managedRoles.getRoleNameFromCanonicalName(role),
              managedRoles.getRoleDescriptionFromCanonicalName(role), )
        output = output + "        }\n"

        output = output + """
#   DisplayCategories are selections of object attributes grouped onto a
#   single display form.  Use them to split up object display into chunks
#   that make sense for you.

defaultGroupDisplayCategories = %s
defaultOrgUnitDisplayCategories = %s
defaultUserDisplayCategories = %s
""" %   ( groupSchema.getDisplayCategories(),
          orgUnitSchema.getDisplayCategories(),
          personSchema.getDisplayCategories(), )

        output = output + """
#   The namingAttribute, or relative distinguished name (RDN) for each
#   object category is the name of the attribute used to name new objects.

defaultGroupNamingAttribute = '%s'
defaultOrgUnitNamingAttribute = '%s'
defaultUserNamingAttribute = '%s'
""" %   ( groupSchema.getNamingAttribute(),
          orgUnitSchema.getNamingAttribute(),
          personSchema.getNamingAttribute(), )

        output = output + """
#   These parameters control the result of searches for objects.  The
#   filter must narrow results to the right category of object, and
#   could include other tests (we filter out inactive employees).  The
#   search keys are the names of the attributes searched for a search
#   string, and are also the attributes returned and displayed.  The
#   searchSortKey is the name of the attribute used to sort the results.
#   The browseFormat is a format string for formatting entries shown in
#   a tree view of the directory.  The browseKeys are attribute names
#   formatted according to the formatString, and the sortKey is for
#   sorting entries under each node of the tree.

defaultGroupSearchFilter = "%s"
defaultGroupSearchKeys = %s
defaultGroupSearchSortKey = '%s'
defaultGroupBrowseFormat = '%s'
defaultGroupBrowseKeys = %s
defaultGroupBrowseSortKey = '%s'
""" %   ( groupSchema.getSearchFilter(),
          groupSchema.getSearchKeys(),
          groupSchema.getSearchSortKey(),
          groupSchema.getBrowseFormat(),
          groupSchema.getBrowseKeys(),
          groupSchema.getBrowseSortKey() )

        output = output + """
defaultOrgUnitSearchFilter = "%s"
defaultOrgUnitSearchKeys = %s
defaultOrgUnitSearchSortKey = '%s'
defaultOrgUnitBrowseFormat = '%s'
defaultOrgUnitBrowseKeys = %s
defaultOrgUnitBrowseSortKey = '%s'
""" %   ( orgUnitSchema.getSearchFilter(),
          orgUnitSchema.getSearchKeys(),
          orgUnitSchema.getSearchSortKey(),
          orgUnitSchema.getBrowseFormat(),
          orgUnitSchema.getBrowseKeys(),
          orgUnitSchema.getBrowseSortKey() )

        output = output + """
defaultUserSearchFilter = "%s"
defaultUserSearchKeys = %s
defaultUserSearchSortKey = '%s'
defaultUserBrowseFormat = '%s'
defaultUserBrowseKeys = %s
defaultUserBrowseSortKey = '%s'
""" %   ( personSchema.getSearchFilter(),
          personSchema.getSearchKeys(),
          personSchema.getSearchSortKey(),
          personSchema.getBrowseFormat(),
          personSchema.getBrowseKeys(),
          personSchema.getBrowseSortKey() )

        output = output + """
#   The pointsToTargets are the options in the pointsTo pull-down
#   menu for managing attributes that says what object category is the
#   value of an attribute whose format is a distinguished name.  This
#   controls the interface to the attribute values, and can't be
#   extended without writing code.

defaultPointsToTargets = %s
""" %   ( personSchema.getPointsToTargets(), )

        output = output + "\ndefaultGroupSchema = {\n"
        output = output + self.getSchemaForSchemaDefaults( groupSchema )

        output = output + "\ndefaultOrgUnitSchema = {\n"
        output = output + self.getSchemaForSchemaDefaults( orgUnitSchema )

        output = output + "\ndefaultUserSchema = {\n"
        output = output + self.getSchemaForSchemaDefaults( personSchema )

        return output

    def iPlanetControllerLDIF( self, output, schema, dn, groupdn, searchFilter, 
        objectCategory ):
        """
        For every 'controller' listed in the schema, create an aci at dn
        to create and delete objects of the given objectCategory for every
        group under groupdn whose cn matches the controller name.
        """
        controllers = []
        roles = self.rolesOfPermission( schema.getPermissionName() )
        for role in roles:
            if role['selected'] == 'SELECTED':
                controllers.append( role['name'] )
        for controller in controllers:
            for controllerDN in self.user.getMatchingGroups( controller, 
                                                             groupdn ):
                output = output + """
add: aci
aci:
 (target="ldap:///%s")
 (targetfilter=%s)
 (version 3.0; acl "%s can create/delete %s";
 allow(add, delete)
 groupdn="ldap:///%s";)
-""" % ( dn, searchFilter, controller, objectCategory+'s', controllerDN )
        return output

    def iPlanetAttributeLDIF( self, output, schema, dn, searchFilter, objectCategory,
                              role, dntype, userdn ):
        #
        # Generate the LDIF statements that describe which attributes 
        # have which permissions for which roles.
	#
	# Attributes controlled by acquired Zope roles need anonymous
	# access at the directory level.  If "role" is the anonymous
	# role 'ANON', then we need to find all of the attributes
	# given Zope role access but not explicitly given anonymous
	# acccess.
        #
        readable = deepcopy(schema.getReadableAttributesByRole( role ))
        writeable = deepcopy(schema.getWriteableAttributesByRole( role ))
	if role == 'ANON':
	    zope_roles = self.aq_parent.aq_parent.valid_roles()
            for r in zope_roles:
	        if r in self._managedRoles.getConfiguredRoleNames():
		    readable = mergeLists( readable,
        			schema.getReadableAttributesByRole( r ))
		    writeable = mergeLists( writeable,
        			schema.getWriteableAttributesByRole( r ))
	    readable.sort()
	    writeable.sort()

        if len( readable ):
            output = output + """
add: aci
aci:
 (target = "ldap:///%s")
 (targetattr = "%s")
 (targetfilter=%s)
 (version 3.0;
 acl "%s can read from %ss";
 allow (read,compare,search)
 (%s = "%s");)
-""" % ( dn, join(readable,' || '), searchFilter, role, objectCategory, 
            dntype, userdn )

        if len( writeable ):
            output = output + """
add: aci
aci:
 (target = "ldap:///%s")
 (targetattr = "%s")
 (targetfilter=%s)
 (version 3.0;
 acl "%s can write to %ss";
 allow (read,compare,search,write)
 (%s = "%s");)
-""" % ( dn, join(writeable,' || '), searchFilter, role, objectCategory, 
            dntype, userdn )

        return output

    def showReportsForm( self, message=None ):
        """
            Given a request for a schema management form, invoke any
            schema updates that may have been called for in the request,
            and then return the appropriate form.
        """

        #   If we got here by a form submit button, then we need to
        #   update the schema first.

        if self.request.has_key('updateSchema'):
            if not self.request.has_key('cancel'):
                message = self.updateSchemaFromRequest()
                self.aq_parent._p_changed = 1
                parent = self.aq_parent
#                self = ZReports( self.user, self.request )
#                self = self.__of__( parent )

        #   Set default form, and render it.  Each form must contain a
        #   hidden tag called 'reportsFormID' whose value, with '_'
        #   prepended, and 'Form' appended, is form rendering DTMLFile
        #   call.

        if not self.request.has_key('reportsFormID'):
            self.request.set('reportsFormID','reportSchemaDoc')

        elif self.request.reportsFormID == 'reportOpenLDAP':
            return self._reportOpenLDAPForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.reportsFormID == 'reportIPlanet':
            return self._reportIPlanetForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.reportsFormID == 'reportSchemaDefaults':
            return self._reportSchemaDefaultsForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.reportsFormID == 'reportSchemaDoc':
            return self._reportSchemaDocForm( 
                        self, self.request, manage_tabs_message = message )

        else:
            message = """
                      The previous form's reportsFormID, %s, does not match a
                      known form display method.
                      """ % self.request.reportsFormID
            return self._schemaManagedForm( self, self.request,
                                            manage_tabs_message = message )

    def updateSchemaFromRequest( self):
        """
            Change the schema according to the REQUEST variables.
        """
        message = None

        #   Invoke a method that matches the name of the submit
        #   button (with '_' prepended).

        if self.request.has_key( 'saveIPlanetACIs' ):
            message = self._saveIPlanetACIs()

        else:
            # You can get here if your form has a single text input
            # field and you press return rather than the form button.
            # Maybe there is a way to figure it out:

            message = "No method to perform requested action on schema."

        return message

    def validRoles(self):
        """
        Return the list of all configured roles, merging lists from
        the separate schema.
        """
        valid = self.user.LDAPDirectory.getPersonSchema().getControllers()
        valid = mergeLists( valid,
                    self.user.LDAPDirectory.getOrgUnitSchema().getControllers())
        valid = mergeLists( valid,
                    self.user.LDAPDirectory.getGroupSchema().getControllers())
        return valid

    def _saveIPlanetACIs( self ):

        # Save the SunONE/iPlanet/Netscape LDIF for the ACIs not
        # generated by this object.

        iPlanetACIs = self.request.get( 'iPlanetACIs', ['',] )
        self.user.LDAPDirectory.setIPlanetACIs( iPlanetACIs )
        return "Updated saved iPlanet ACIs."

    def zopeRoleNeedsANON(self, anonymousRoleName, roles):
	# If one of 'roles' is an acquired Zope role, then at the
	# directory level anonymous access must be allowed.  If 
	# anonymous access is not in 'roles,' then return the list of
	# acquired roles, so we can warn the user.  We add ANON when 
	# writing the security assertions in the ZReports class, not here.

	zope_roles = self.aq_parent.aq_parent.valid_roles()
	if anonymousRoleName in roles:
	    return []

	acquiredRoles = []
	for role in roles:
	    if role in zope_roles:
		acquiredRoles.append(role)

	return acquiredRoles

