"""

    Module: $RCSfile: ZSchema.py,v $
    Author: $Author: payne $
    Date:   $Date: 2003/12/16 00:35:12 $

    The ZSchema class is the mutable schema object class.  It holds
    an AuthenticatedUser instance, a ManagedSchema class instance,
    a ManagedRoles instance, the REQUEST, and contains methods used 
    on the various faces of the Schema tab.

    Schema management is split into three parts:  the schemas for
    'person', 'group', and 'orgUnit' types of object are managed
    separately.  The Schema tab provides a way to select which
    type of object schema is being managed.  The schema tab actually 
    consists of a number of forms:

      o A form for selecting the objectClass list to be associated
        with the object category.  The "Selected Object Classes" list
        contains the object classes expected of all objects of this
        category in the directory, and is used whenever a new object 
        is created.  The "Available Object Classes" list contains
        object classes values that might be held by objects in the
        directory.

      o A form for selecting the attributes to be managed from the list
        of all attributes available in the available and selected
        object classes, selected on the form just described.

      o A form for managing a list of "Display Categories."  A display
        category is a list of attributes to be managed together on a
        form named for the display category.  By creating display
        categories, and grouping attributes under them, you can
        configure the user interface for managing an object in LDAP.

      o A form for managing lists of roles.  This list of roles is used
        for configuring the individual managed attributes, when
        specifying who has read and write access to the attribute.
        There is a list of special "Permanent" roles, not associated
        with LDAP groups, which can be augmented with "Configured"
        roles, which do correspond with LDAP groups.

      o A form for editing the schema of individual managed attributes,
        allowing the manager to supply properties like a friendly name,
        and a list of valid values, and to specify which Zope roles
        have read and write access to the attribute.

      o A form for specifying how to search for objects of this category,
        and how to display the results.  This is done by specifying an 
        LDAP search filter for selecting objects to be queried, and then 
        a list of attributes, whose values are searched for a match to
        a search string.  This list of attributes is also used for
        displaying the results as a table, with control over the column
        headings and the column to be used for sorting.

      o A form for specifying which roles can create and delete objects
        in LDAP.

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

"""
__version__ = '$Revision: 1.85 $'[11:-2]

from AccessControl import ClassSecurityInfo
from AccessControl.Permission import Permission
from AccessControl.Role import RoleManager
from Acquisition import Implicit
from Globals import DTMLFile
from string import join, lower

from LDAPExceptions import REQUIRED_ATTRIBUTE, NO_SUCH_ATTRIBUTE
from LDAPUtilities import _ldap, DNsAreEquivalent, \
    list1ContainsList2, mergeLists, TRUE, FALSE
from LDIFObjectParser import LDIFObjectParser
from LDIFAttributeParser import LDIFAttributeParser
from ManagedAttributeSchema import ManagedAttributeSchema
from ZRoles import ZRoles

class ZSchema( Implicit ):

    security = ClassSecurityInfo()
    security.declareProtected('LDAP Manage Schema', 
        'manage_schemaFormSelection',
        )

    manage_schemaFormSelection = DTMLFile('dtml/schemaFormSelection', globals())
    manage_schemaSearchSelection = DTMLFile('dtml/schemaSearchSelection', globals())

    #   DTML templates for the various "faces" of the Schema tab.
    #   Each template's forms must contain a hidden variable that
    #   identifies the form.  The name is 'schemaFormID', and the
    #   value must match one of the following after prepending an
    #   underscore, and appending 'Form', as in:
    #
    #     <input type=hidden name=schemaFormID value=schemaAttributes>
 
    _schemaAttributesForm = DTMLFile('dtml/schemaAttributes', globals())
    _schemaAttrsByCategoryForm = DTMLFile('dtml/schemaAttrsByCategory',
                                          globals())
    _schemaAttrsByRoleForm = DTMLFile('dtml/schemaAttrsByRole', globals())
    _schemaBrowsingForm = DTMLFile('dtml/schemaBrowsing', globals(),
                                    management_view='Searching & Browsing')
    _schemaCategoriesForm = DTMLFile('dtml/schemaCategories', globals())
    _schemaControllersForm = DTMLFile('dtml/schemaControllers', globals())
    _schemaManagedForm = DTMLFile('dtml/schemaManaged', globals())
    _schemaNamingAttrForm = DTMLFile('dtml/schemaNamingAttribute', globals())
    _schemaObjectClassesForm = DTMLFile('dtml/schemaObjectClasses',
                                        globals())
    _schemaOrgUnitIndicatorForm = DTMLFile('dtml/schemaOrgUnitIndicator', 
                                        globals())
    _schemaRenameCategoriesForm = DTMLFile('dtml/schemaRenameCategories', 
                                           globals())
    _schemaSearchFilterForm = DTMLFile('dtml/schemaSearchFilter',globals(),
                                    management_view='Searching & Browsing')
    _schemaSearchKeysForm = DTMLFile('dtml/schemaSearchKeys',globals(),
                                    management_view='Searching & Browsing')

    def __init__(self, user, REQUEST ):

        #   Update the Zope schema object with values from the LDAP 
        #   schema (and save the list of available LDAP object classes).  
        
        self.user = user
        self.request = REQUEST

        if not self.request.has_key('objectCategory'):
            self.request.set('objectCategory','person')
        category = lower(self.request['objectCategory'])

        if category == 'group':
            self._schema = user.getLDAPDirectory().getGroupSchema()
        elif category == 'person':
            self._schema = user.getLDAPDirectory().getPersonSchema()
        elif category == 'orgunit':
            self._schema = user.getLDAPDirectory().getOrgUnitSchema()
        else:
            assert FALSE            

        self._ldapObjects = self.updateSchemaFromLDAP( user )
        self._managedRoles = user.getLDAPDirectory().getManagedRoles()

    def checkObjectClassCompleteness( self, objectClasses ):
        """
            Given a list of objectClasses, check that the list is
            complete, in the sense that all of the superclasses
            for each item in the list are also in the list.
        """
        for className in objectClasses:
            superClasses = getLDAPSuperClassNames( className,
                                                   self._ldapObjects )
            for item in superClasses:
                if item not in objectClasses:
                    return FALSE
        return TRUE

    def DNsAreEquivalent( self, dn1, dn2):
        """
           Wrapper for LDAPUtilities.DNsAreEquivalent
        """
        return DNsAreEquivalent( dn1, dn2 )

    def getManagedRoles(self):
        """
            Returns a managed roles object.
        """
        return self._managedRoles

    def getManagedSchema(self):
        """
        Returns a managed schema object,
        """
        return self._schema

    def getSchemaPermission(self):
        """
        Returns the name of the permission for managing the schema
        """
        return self.manageSchemaPermission

    def getUnmanagedObjectClasses( self ):
        """
            Returns a list of names of all of the object classes in 
            the LDAP schema, except for the ones already managed 
            (i.e., either in available or selected object class names).
        """
        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
        managed = mergeLists( objectClassSchema.getValidValues(),
                              objectClassSchema.getRequiredValues() )
        unmanaged = filter( lambda i,m=managed: i not in m,
                            self._ldapObjects.keys() )
        unmanaged.sort()
        return unmanaged

    def showSchemaForm( self, message ):
        """
            Given a request for a schema management form, invoke any
            schema updates that may have been called for in the request,
            and then return the appropriate form.
        """

        #   If we got here by a form submit button, then we need to
        #   update the schema first.

        if self.request.has_key('updateSchema'):
            if not self.request.has_key('cancel'):
                message = self.updateSchemaFromRequest()
                self.aq_parent._p_changed = 1
                parent = self.aq_parent
                self = ZSchema( self.user, self.request )
                self = self.__of__( parent )

        #   Set default form, and render it.  Each form must contain a
        #   hidden tag called 'schemaFormID' whose value, with '_'
        #   prepended, and 'Form' appended, is form rendering DTMLFile
        #   call.

        if not self.request.has_key('schemaFormID'):
            self.request.set('schemaFormID','schemaObjectClasses')

        elif self.request.schemaFormID == 'schemaObjectClasses':
            return self._schemaObjectClassesForm(
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaAttributes':
            return self._schemaAttributesForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaAttrsByCategory':
            return self._schemaAttrsByCategoryForm(
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaAttrsByRole':
            return self._schemaAttrsByRoleForm(
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaManaged':
            return self._schemaManagedForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaNamingAttribute':
            return self._schemaNamingAttrForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaCategories':
            return self._schemaCategoriesForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaControllers':
            return self._schemaControllersForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaRenameCategories':
            return self._schemaRenameCategoriesForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaSearchFilter':
            return self._schemaSearchFilterForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaSearchKeys':
            return self._schemaSearchKeysForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaBrowsing':
            return self._schemaBrowsingForm( 
                        self, self.request, manage_tabs_message = message )

        elif self.request.schemaFormID == 'schemaOrgUnitIndicator':
            return self._schemaOrgUnitIndicatorForm( 
                        self, self.request, manage_tabs_message = message )

        else:
            message = """
                      The previous form's schemaFormID, %s, does not match a
                      known form display method.
                      """ % self.request.schemaFormID
            return self._schemaManagedForm( self, self.request,
                                            manage_tabs_message = message )

    def nextAttributeLink( self ):
        """
            Construct a link to the next managed attribute.
        """
        managed = self._schema.getManagedAttributeNames()
        ldapName = self.request.get('ldapName')
        attributeIndex = managed.index( ldapName )

        if attributeIndex == len(managed)-1:
            return "&nbsp;"
        else:
            link = "<a href=\"%s?objectCategory=%s" % \
             ( self.request.get('URL0'), self.request.get('objectCategory')) + \
              "&schemaFormID=%s&ldapName=%s\">" % \
             ( self.request.get('schemaFormID'), managed[ attributeIndex+1]) + \
                   "Next</a>"
            return link

    def nextRoleLink( self ):
        """
            Construct a link to the next role.
        """
        roles = self._managedRoles.getCanonicalRoleNames()
        roleName = self.request.roleName
        roleIndex = roles.index( roleName )

        if roleIndex == len(roles)-1:
            return "&nbsp;"
        else:
            link = "<a href=\"%s?objectCategory=%s" % \
                   ( self.request['URL0'], self.request['objectCategory']) + \
                   "&schemaFormID=%s&roleName=%s\">" % \
                   ( self.request['schemaFormID'], roles[ roleIndex+1]) + \
                   "Next&nbsp;</a>"
            return link

    def prevAttributeLink( self):
        """
            Construct a link to the previous managed attribute.
        """
        managed = self._schema.getManagedAttributeNames()
        ldapName = self.request.get('ldapName')
        attributeIndex = managed.index( ldapName )

        if attributeIndex == 0:
            return "&nbsp;"
        else:
            link = "<a href=\"%s?objectCategory=%s" % \
                   ( self.request.get('URL0'), self.request.get('objectCategory')) + \
                   "&schemaFormID=%s&ldapName=%s\">" % \
                   ( self.request.get('schemaFormID'), managed[ attributeIndex-1]) + \
                   "Previous</a>"
            return link

    def prevRoleLink( self ):
        """
            Construct a link to the previous role.
        """
        roles = self._managedRoles.getCanonicalRoleNames()
        roleName = self.request.roleName
        roleIndex = roles.index( roleName )

        if roleIndex == 0:
            return "&nbsp;"
        else:
            link = "<a href=\"%s?objectCategory=%s" % \
                   ( self.request['URL0'], self.request['objectCategory']) + \
                   "&schemaFormID=%s&roleName=%s\">" % \
                   ( self.request['schemaFormID'], roles[ roleIndex-1]) + \
                   "&nbsp;Previous</a>"
            return link

    def updateSchemaFromLDAP( self, user ):
        """
            Retrieves the schema from the LDAP directory server
            and updates the persisted schema collection.
        """
        # Search for the subschemaSubentry attribute of the root
        # object, which gives the DN of the schema.  Get the schema
        # objectClasses and attributesTypes, each of which is a list
        # of LDIF expressions, and parse them into dictionaries of
        # parsed objects, which have the LDIF data as attributes.

        LDAPschemaDN = user.search( "", _ldap.SCOPE_BASE, "(objectclass=*)",
            ['subschemaSubentry',])[0][1]['subschemaSubentry'][0]

        LDAPschema = user.search( LDAPschemaDN, _ldap.SCOPE_BASE,
            "(objectclass=*)", ["objectClasses","attributeTypes"] )[0][1]

        ldapObjects = {}
        for ldifExpression in LDAPschema['objectClasses']:
            parsedObject = LDIFObjectParser( ldifExpression )
            ldapObjects[lower(parsedObject.name)] = parsedObject

        #   Given the list of objectClass definitions, build up a list
        #   of attribute names available to this category of object.
        #   The required object classes are the ones expected of every
        #   object of this category in the directory.  New objects of
        #   this category will be created with objectClass set to the
        #   list of required object classes.  The set of attributes
        #   required by the required object classes will be marked as
        #   required.
        #
        #   The available object classes are those that might be
        #   present in some objects of this category in the directory.
        #   New objects of this category are not given these object
        #   class values.  The set of attributes required by the
        #   available object classes will NOT be marked as required.
	#
	#   Only objectClasses known in the schema can be required or
	#   available.  Filter out anything else.

        requiredLDAPAttrNames = []
        optionalLDAPAttrNames = []

        requiredObjectClasses = \
            self._schema.getAttributeSchema('objectClass').getRequiredValues()
	requiredObjectClasses = \
	    filter( lambda i,keys=ldapObjects.keys(): lower(i) in keys, 
	            requiredObjectClasses )
	self._schema.getAttributeSchema('objectClass').setRequiredValues(
		requiredObjectClasses )

        validObjectClasses = \
            self._schema.getAttributeSchema('objectClass').getValidValues()
	validObjectClasses = \
	    filter( lambda i,keys=ldapObjects.keys(): lower(i) in keys, 
	            validObjectClasses )
	self._schema.getAttributeSchema('objectClass').setValidValues(
		validObjectClasses )

        for objectClass in requiredObjectClasses:

            required, optional = getLDAPAttributeNames(objectClass, ldapObjects)
            requiredLDAPAttrNames = mergeLists(requiredLDAPAttrNames, required)
            optionalLDAPAttrNames = mergeLists(optionalLDAPAttrNames, optional)

        availableObjectClasses = \
            self._schema.getAttributeSchema('objectClass').getAvailableValues()

        for objectClass in availableObjectClasses:

            required, optional = getLDAPAttributeNames(objectClass, ldapObjects)
            optionalLDAPAttrNames = mergeLists(optionalLDAPAttrNames, required)
            optionalLDAPAttrNames = mergeLists(optionalLDAPAttrNames, optional)

        optionalLDAPAttrNames = filter( lambda i,r=requiredLDAPAttrNames:
                                        i not in r,
                                        optionalLDAPAttrNames )
                                  
        allowedLDAPAttrNames = mergeLists(requiredLDAPAttrNames,
                                          optionalLDAPAttrNames )

        allowedLDAPAttrNames.sort()
        self._schema._allowedLDAPAttrNames = allowedLDAPAttrNames

        # Now parse the attribute definitions.  Because attributes
        # can be subclassed from other attributes, the ldapAttributes
        # dictionary may contain parsed objects that have a specified
        # superior but no other attributes.  For attributes with
        # a superior, find the superior attribute parsed object,
        # and passing it back to the target attribute's inheritsFrom
        # method.

        ldapAttributes = {}
        for ldifExpression in LDAPschema['attributeTypes']:
            parsedObject = LDIFAttributeParser( ldifExpression )
            ldapAttributes[lower(parsedObject.name)] = parsedObject
            if parsedObject.aliases:
                for alias in parsedObject.aliases:
                    ldapAttributes[lower(alias)] = parsedObject

        for ldapAttrName in self._schema._allowedLDAPAttrNames:
            parsedAttr = ldapAttributes[lower(ldapAttrName)]
            if parsedAttr.sup:
                superiorAttrName = parsedAttr.sup
                while superiorAttrName:
                    superiorAttr = ldapAttributes[lower(superiorAttrName)]
                    parsedAttr.inheritsFrom( superiorAttr )
                    superiorAttrName = superiorAttr.sup

        #   Update the schema attribute collection.  Get the metadata
        #   from items already in the attribute collection, and update
        #   that with metadata from the LDAP schema.  New items only
        #   get LDAP schema metadata, but they are set to be unmanaged,
        #   and to be readable and writeable by the Controllers for
        #   this category of object.  Update the 'required' metadata
        #   item, and then create a new ManagedAttributeSchema object
        #   and put it in the attribute collection.  Attribute
        #   schemas no longer allowed, because the objectClass that 
        #   defined it has been dropped, are deleted.

        for ldapName in allowedLDAPAttrNames:

            if ldapName in self._schema._attributeCollection.keys():
                metadata = \
                    self._schema.getAttributeSchema(ldapName).getMetadata()
                metadata.update(
                    ldapAttributes[lower(ldapName)].getMetadata() )
            else:
                metadata = ldapAttributes[lower(ldapName)].getMetadata()
                metadata = { 'managed': FALSE,
                             'readers': self._schema._Controllers,
                             'writers': self._schema._Controllers, }

            if ldapName in requiredLDAPAttrNames:
                metadata['required'] = TRUE
                metadata['managed'] = TRUE
            else:
                metadata['required'] = FALSE

            managedAttribute = ManagedAttributeSchema( ldapName, metadata )
            self._schema._attributeCollection[ ldapName ] = managedAttribute

        for ldapName in self._schema._attributeCollection.keys():
            if not ldapName in allowedLDAPAttrNames:
                del self._schema._attributeCollection[ldapName]                

        #   Force the namingAttribute to be required (and managed).

        namingAttribute = self._schema.getNamingAttribute()
        namingAttributeAttr = self._schema.getAttributeSchema(namingAttribute)
        namingAttributeAttr.setRequired(TRUE)
        namingAttributeAttr.setManaged(TRUE)

        #   The list of available LDAP objects is used to check new
        #   values for the 'objectClass' attribute, so we return it
        #   for ZSchema use.  But the ManagedSchema object doesn't 
        #   need it.

        return ldapObjects
    
    def updateSchemaFromRequest( self):
        """
            Change the schema according to the REQUEST variables.
        """
        message = None

        #   Invoke a method that matches the name of the submit
        #   button (with '_' prepended).

        if self.request.has_key( 'saveAttributeSchema' ):
            message = self._saveAttributeSchema()

        elif self.request.has_key( 'setNamingAttribute' ):
            message = self._setNamingAttribute()

        elif self.request.has_key( 'moveAttributeToAvailable' ):
            message = self._moveAttributeToAvailable()

        elif self.request.has_key( 'moveAttributeToManaged' ):
            message = self._moveAttributeToManaged()

        elif self.request.has_key( 'moveObjectClassesToSelected' ):
            message = self._moveObjectClassesToSelected()

        elif self.request.has_key( 'moveObjectClassesToAvailable' ):
            message = self._moveObjectClassesToAvailable()

        elif self.request.has_key( 'dropObjectClasses' ):
            message = self._dropObjectClasses()

        elif self.request.has_key( 'addAvailableObjectClasses' ):
            message = self._addAvailableObjectClasses()

        elif self.request.has_key( 'changeDisplayCategories' ):
            message = self._changeDisplayCategories()

        elif self.request.has_key( 'renameDisplayCategories' ):
            message = self._renameDisplayCategories()

        elif self.request.has_key( 'saveAttrsByCategory' ):
            message = self._saveAttrsByCategory()

        elif self.request.has_key( 'saveAttrsByRole' ):
            message = self._saveAttrsByRole()

        elif self.request.has_key( 'saveBrowse' ):
            message = self._saveBrowse()

        elif self.request.has_key( 'addControllers' ):
            message = self._addControllers()

        elif self.request.has_key( 'deleteControllers' ):
            message = self._deleteControllers()

        elif self.request.has_key( 'addBrowseKeys' ):
            message = self._addBrowseKeys()

        elif self.request.has_key( 'addSearchKeys' ):
            message = self._addSearchKeys()

        elif self.request.has_key( 'deleteBrowseKeys' ):
            message = self._deleteBrowseKeys()

        elif self.request.has_key( 'deleteSearchKeys' ):
            message = self._deleteSearchKeys()

        elif self.request.has_key( 'searchFilterSave' ):
            message = self._searchFilterSave()

        elif self.request.has_key( 'searchKeysSave' ):
            message = self._searchKeysSave()

        elif self.request.has_key( 'searchOrgUnitIndicatorSave' ):
            message = self._searchOrgUnitIndicatorSave()

        elif self.request.has_key( 'searchBrowsingSave' ):
            message = self._searchBrowsingSave()

        elif self.request.has_key( 'firstBrowseKeys' ):
            message = self._firstBrowseKeys()

        elif self.request.has_key( 'firstSearchKeys' ):
            message = self._firstSearchKeys()

        else:
            # You can get here if your form has a single text input
            # field and you press return rather than the form button.
            # Maybe there is a way to figure it out:

            if self.request.get('schemaFormID','') == 'schemaSearchFilter':
                message = self._searchFilterSave()
            else:
                message = "No method to perform requested action on schema."

        return message

    def validRoles(self):
        """
        Return the list of all configured roles, merging lists from
        the separate schema.
        """
        valid = self.user.LDAPDirectory.getPersonSchema().getControllers()
        valid = mergeLists( valid,
                    self.user.LDAPDirectory.getOrgUnitSchema().getControllers())
        valid = mergeLists( valid,
                    self.user.LDAPDirectory.getGroupSchema().getControllers())
        return valid

    def _addAvailableObjectClasses( self ):

        # A method for adding items to the available object class
        # list.  Each input name is checked against the LDAP
        # schema to make sure it exists.

        newClasses = self.request.get('newClasses',None)
        if newClasses is None:
            return

        schema = self._schema.getAttributeSchema( 'objectClass' )
        valid = schema.getValidValues()
        added = []
        bogus = []
        super = []

        for newClass in newClasses:
            if self._ldapObjects.has_key( lower(newClass) ):

                # Get canonical name.

                name = self._ldapObjects[lower(newClass)].name
                schema.addValidValues( name )
                valid.append( name )
                added.append( name )

                # Do any superclasses also need to be added?

                supers = getLDAPSuperClassNames( name,
                                                      self._ldapObjects )
                supersToAdd = filter( lambda i,v=valid: i not in v, supers )
                for superClass in supersToAdd:
                    name = self._ldapObjects[lower(superClass)].name
                    schema.addValidValues( name )
                    valid.append( name )
                    super.append( name )

            else:
                bogus.append( name )

        message = ""
        if len( added ) > 0:
            message = message + \
                      "Added object classes %s.  " % ( str( added ) )
        if len( super ) > 0:
            message = message + \
                      "Added missing superClasses %s.  " % ( str( super ) )
        if len( bogus ) > 0:
            message = message + \
                      "No matching LDAP classes for %s." % ( str( bogus ) )
        return message

    def _addBrowseKeys( self ):
        """
            Extend the list of browse keys with values from the REQUEST.
        """
        allowedKeys = self._schema.getManagedAttributeNames()
        newBrowseKeys = self.request.get( 'browseKeys', [] )
        newBrowseKeys = filter( lambda i,ids=allowedKeys: i in ids, 
                                newBrowseKeys )
        self._schema.addBrowseKeys( newBrowseKeys )
	self._setBrowseFormat()
	self._setBrowseSortKey()
	if newBrowseKeys:
            return "Added attributes %s" % newBrowseKeys
	else:
	    return ""

    def _addControllers( self ):

        # A method for adding the the controllers list

        ids = self.request.get( 'newRoles', [] )

	# A controller needs write access to the namingAttribute to
	# permit renaming an entry in OpenLDAP.

	namingAttribute = self._schema.getNamingAttribute()
	namingAttrSchema = self._schema.getAttributeSchema(namingAttribute)

        zopeRoleSearchBase = \
            self.user.getLDAPDirectory().getZopeRoleSearchBase()
        for item in ids:
            self._schema.addController(item)
	    self._schema.getAttributeSchema(namingAttribute).addWriters(item)
        self._assignRolesToPermission()
        return "Added roles."

    def _addSearchKeys( self ):
        """
            Extend the list of search keys with values from the REQUEST.
        """
        self._searchKeysSave()
        allowedKeys = self._schema.getManagedAttributeNames()
        newSearchKeys = self.request.get( 'searchKeys', [] )
        newSearchKeys = filter( lambda i,ids=allowedKeys: i in ids, 
                                newSearchKeys )
	if newSearchKeys:
            self._schema.addSearchKeys( newSearchKeys )
	    return "Added search key(s)."
	else:
	    return ""

    def _assignRolesToPermission( self, roles=(), permissionName=None ):
        """
	Assign permission to view tab for this type of object.
        """
        for role in roles:
            RoleManager._addRole( self.aq_parent, role )
	
        # Ensure that controllers for this type of object have the
	# permission to view the tab for managing this type of object.

        controllers = self._schema.getControllers()

        for role in controllers:
            RoleManager._addRole( self.aq_parent, role )

	# Assign the permission.  Use the permission name for this
	# schema if None was supplied.

	if permissionName is None:
	    permissionName = self._schema.getPermissionName()

	addedRoles = list(roles) + list(controllers)
        RoleManager.manage_permission( self.aq_parent, permissionName,
                                       addedRoles, acquire=TRUE )

    def _browseKeysSave(self):

        # Save browse keys specified in REQUEST

        newSortKey = self.request.get( 'browseSortKey' )
        if newSortKey in self._schema.getBrowseKeys():
            self._schema.setBrowseSortKey( newSortKey )
        elif self._schema.getBrowseKeys():
            self._schema.setBrowseSortKey(self._schema.getBrowseKeys()[0])
        else:
            self._schema.setBrowseSortKey(None)

        return "Saved changes."

    def _changeDisplayCategories( self ):

        # A method for updating the display categories list

        option = self.request.get( 'changeDisplayCategories' )
        ids = self.request.get( 'ids', [] )

        if option == " Delete " and ids:
            for item in ids:
                if item in self._schema.getDisplayCategories():
                    self._schema.delDisplayCategory( item )
            message = "Deleted object categories."

        elif option == " First " and ids:
            displayCategories = filter( lambda i,ids=ids: i in ids,
                                        self._schema.getDisplayCategories())+\
                                filter( lambda i,ids=ids: i not in ids,
                                        self._schema.getDisplayCategories())
            self._schema.setDisplayCategories( displayCategories )
            message = "Re-ordered display categories."

        elif option == " Add ":
            newCategory = self.request.get( 'newCategory', None )
            if newCategory:
                self._schema.addDisplayCategory( newCategory )
                message = "Added object category."
	    else:
		message = ""

        else:
            message = ""

        return message

    def _deleteBrowseKeys( self):

        # Delete browse keys

        ids = self.request.get( 'ids' )
	if ids:
            browseKeys = self._schema.getBrowseKeys()
            browseKeys = filter( lambda i,ids=ids: i not in ids, browseKeys )
            self._schema.setBrowseKeys( browseKeys )
	    self._setBrowseFormat()
	    self._setBrowseSortKey()

            return "Deleted attributes %s." % ids

	else:
	    self._setBrowseFormat()
	    self._setBrowseSortKey()
	    return ""

    def _deleteControllers( self ):

        # A method for dropping items from the controllers list

	namingAttribute = self._schema.getNamingAttribute()
	namingAttrSchema = self._schema.getAttributeSchema(namingAttribute)

        ids = self.request.get( 'ids', [] )
        for item in ids:
            self._schema.delController( item )
	    self._schema.getAttributeSchema(namingAttribute).delWriters(item)
        return "Deleted roles."

    def _deleteSearchKeys( self):

        # Delete search keys

        ids = self.request.get( 'ids' )
	if ids:
            searchKeys = self._schema.getSearchKeys()
            searchKeys = filter( lambda i,ids=ids: i not in ids, searchKeys )
            self._schema.setSearchKeys( searchKeys )

            self._searchKeysSave()

            return "Deleted search keys %s." % ids
	else:
            self._searchKeysSave()
	    return ""

    def _dropAvailableObjectClasses( self, dropClasses ):

        # Drop the specifed object classes from the available
        # object classes list.

        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
        objectClassSchema.delValidValues( dropClasses )
        return "Dropped specified objectClasses."

    def _dropObjectClasses( self ):

        # Drop object classes from both the available and selected
        # object class lists.  

        message = ""
        availClasses = self.request.get('availClasses',[])
        if len( availClasses ) > 0:
            message = message + \
                      self._dropSelectedObjectClasses( availClasses )

        selectClasses = self.request.get('selectClasses',[])
        if len( selectClasses ) > 0:
            message = message + \
                      self._dropAvailableObjectClasses( selectClasses )

        return message

    def _dropSelectedObjectClasses( self, dropClasses ):

        # Drop the specified classes from the selected object classes
        # list.  But you can't drop an object class subclassed by one
        # of the classes remaining in the list.  In that case, the
        # drop fails.  Compute what the new selected object classes list 
        # would be, and see if it is complete.

        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
        requiredValues = objectClassSchema.getRequiredValues()
        for item in dropClasses:
            requiredValues.remove( item )
        if not self.checkObjectClassCompleteness( requiredValues ):

            # What classes would be stranded with a missing superclass?

            stranded = []
            for item in requiredValues:
                supers = getLDAPSuperClassNames( item,
                                                      self._ldapObjects )
                if not list1ContainsList2( requiredValues, supers ):
                    stranded.append( item )

            message = """
                      The requested object class(es), %s, could not 
                      be moved because the list contains superclasses
                      for classes %s.  Only mixin and lowest level 
                      classes can be moved.  Either move all of these
                      classes, or pick items from the second list.
                      """ % ( str(dropClasses), str(stranded) )

            return message

        #   The REQUEST is missing the 'validValues' and 'requiredValues'
        #   keys that any other attribute schema update would have.
        #   Supply them, and then call the _saveAttributeSchema method.

        objectClassSchema.setRequiredValues( requiredValues )

        message = self._saveAttributeSchema()
        return "Dropped specified objectClasses.  " + str(message)

    def _firstBrowseKeys( self):

        self._browseKeysSave()
	self._setBrowseFormat()

        # Move the select keys to the head of the list

        browseKeys = self._schema.getBrowseKeys()
        ids = self.request.get( 'ids', [] )
        browseKeys = filter(lambda i,ids=ids: i in ids, browseKeys ) \
                     + filter(lambda i,ids=ids: i not in ids, browseKeys )
	if browseKeys:
            self._schema.setBrowseKeys( browseKeys )
            return "Reordered browse keys."
	else:
	    return ""

    def _firstSearchKeys( self):

        self._searchKeysSave()

        # Move the select keys to the head of the list

        searchKeys = self._schema.getSearchKeys()
        ids = self.request.get( 'ids', [] )
        searchKeys = filter(lambda i,ids=ids: i in ids, searchKeys ) \
                     + filter(lambda i,ids=ids: i not in ids, searchKeys )
        self._schema.setSearchKeys( searchKeys )
	if ids:
            return "Reordered search keys."
	else:
	    return ""

    def _moveAttributeToAvailable(self):

        # A method for making an attribute "available," rather than
        # "managed."

        selectedAttrs = self.request.get('unmanageAttrs',[])
        for item in selectedAttrs:
            self._schema.unmanageAttributeSchema( item )

        return "Updated attribute lists."

    def _moveAttributeToManaged(self):

        # A method for making an attribute "managed," rather than
        # "available."

        selectedAttrs = self.request.get('manageAttrs',[])
        for item in selectedAttrs:
            self._schema.manageAttributeSchema( item )

        return "Updated attribute lists."

    def _moveObjectClassesToAvailable(self):

        # A method for making an object class value "available,"
        # rather than "selected."  But you can't unselect an
        # object class subclassed by some other object in the
        # required list.

        availClasses = self.request.get('availClasses',[])
        message = self._dropSelectedObjectClasses(availClasses)
        if message[0:31] == "Dropped specified objectClasses.":
            # i.e., the drop worked
            objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )
            objectClassSchema.addValidValues( availClasses )
            message = self._saveAttributeSchema()
            return "Moved specified objectClasses.  " + str(message)

        else:
            return message

    def _moveObjectClassesToSelected(self):

        # A method for making an object class value "selected,"
        # rather than "available."  In addition to any specified
        # classes, be sure to move their superclasses, too.

        selectClasses = self.request.get('selectClasses',[])
        supers = []
        for item in selectClasses:
            supers = mergeLists( supers,
                                 getLDAPSuperClassNames( item,
                                                         self._ldapObjects ) 
                               )
        
        objectClassSchema = self._schema.getAttributeSchema( 'objectClass' )

        requiredValues = mergeLists( objectClassSchema.getRequiredValues(),
                                     selectClasses )

        # Now reassign any superclasses that also need to move

        supersToAdd = filter( lambda i,r=requiredValues: i not in r, supers )
        if len( supersToAdd ) > 0:
            requiredValues = mergeLists( requiredValues, supersToAdd )
            message = "  Added superclasses to the required class list."

        objectClassSchema.setRequiredValues( requiredValues )

        message = self._saveAttributeSchema()

        return "Updated object class lists.  " + str(message)

    def _renameDisplayCategories(self):
        """
            Rename display categories.
        """
        #   If we got here from the _schemaRenameCategoriesForm, then 
        #   perform the rename.

        if self.request.has_key('haveNewNames'):
            ids = self.request.get( 'ids', [] )
            new_ids = self.request.get( 'new_ids', [] )
            displayCategories = self._schema.getDisplayCategories()

            # Skip empty new_ids

            while '' in new_ids:
                i = new_ids.index('')
                ids.pop(i)
                new_ids.pop(i)

            for i in range( 0, len( ids ) ):
                oldName = ids[ i ]
                newName = new_ids[ i ]

                #   Easy: change the list ...

                displayCategories[ displayCategories.index( oldName ) ] = \
                    newName

                #   Hard: change all of the attribute objects.

                for attributeSchema in self._schema._attributeCollection.values():
                    if oldName in attributeSchema.getDisplayCategories():
                        attributeSchema.delDisplayCategories( oldName )
                        attributeSchema.addDisplayCategories( newName )

            self._schema.setDisplayCategories( displayCategories )
            return "Renamed display categories"

        #   Otherwise, set REQUEST variables so that the
        #   _schemaRenameCategoriesForm gets displayed.

        else:
            ids = self.request.get( 'ids', [] )
	    if ids:
                self.request.set('schemaFormID', 'schemaRenameCategories')
                del self.request.form['updateSchema']

            return ""

    def _saveAttributeSchema(self):
        """
        Save an updated attribute schema.
        """
        ldapName = self.request['ldapName']
        schema = self._schema.getAttributeSchema( ldapName )

        if not ldapName == 'objectClass':

            schema.setValidValues( self.request.get('validValues') )
            schema.setRequiredValues( self.request.get('requiredValues',[]) )

        schema.setDescription( self.request.get('description', '') )
        schema.setDisplayCategories( self.request.get('displayCategories',[]) )
        schema.setFriendlyName( self.request.get('friendlyName', '') )
        schema.setManaged( self.request.get('managed', FALSE ) )
        schema.setPointsTo( self.request.get('pointsTo', '') )
        schema.setRequired( self.request.get('required', FALSE ) )
        schema.setSyntaxByOID( self.request.get('syntax', None) )

        readers = self.request.get('readers',[])
        writers = self.request.get('writers',[])
        schema.setReaders( readers )
        schema.setWriters( writers )

	# If there are acquired Zope roles among the readers or writers,
	# then, at the directory level, the 'ANON' role must apply, since
	# the directory doesn't know anything about Zope roles.  If 
	# 'ANON' has not been specified, then don't set the acquired
	# role, and print an explanatory message.

	anonymousRoleName = \
	    self._managedRoles.getRoleNameFromCanonicalName( 'ANON' )
	zopeReaders = self.zopeRoleNeedsANON( anonymousRoleName, readers )
        zopeWriters = self.zopeRoleNeedsANON( anonymousRoleName, writers )

	if zopeReaders and zopeWriters:
	    return "Zope role(s) %s can have read and write access only if the %s role has read and write access.  Access controls written under the Reports tab will allow anonymous access to this attribute.  Decide now whether this is a good idea (in the case of writing this is probably a bad idea)." % \
              ( join( mergeLists( zopeReaders, zopeWriters ), ', '),
                anonymousRoleName, )
	elif zopeReaders:
	    return "Zope role(s) %s can have read access only if the %s role has read access.  Access controls written under the Reports tab will allow anonymous access to this attribute.  Decide now whether this is a good idea." % \
              ( join( zopeReaders, ', '), anonymousRoleName, )
	elif zopeWriters:
	    return "Zope role(s) %s can have write access only if the %s role has write access.  Access controls written under the Reports tab will allow anonymous access to this attribute.  This is probably not a good idea, so consider changing the list of Writers." % \
              ( join( zopeWriters, ', '), anonymousRoleName, )
	else:
	    return "Updated schema attributes."

    def _saveAttrsByCategory(self):
        """
        Save results of the schemaAttrsByCategory form.
        """
        displayCategory = self.request[ 'displayCategory' ]
        ids = self.request['ids']

        for ldapName in self._schema.getManagedAttributeNames():

            schema = self._schema.getAttributeSchema( ldapName )
            if ldapName in ids:
                schema.addDisplayCategories( displayCategory )
            elif displayCategory in schema.getDisplayCategories():
                schema.delDisplayCategories( displayCategory )

        return "Saved display category \"%s\"'s presence or absence in each managed attribute." % ( displayCategory )

    def _saveAttrsByRole(self):
        """
            Save results of the schemaAttrsByRole form.
        """
        roleName = self.request[ 'roleName' ]
        readIds = self.request.get('readIds',[])
        writeIds = self.request.get('writeIds',[])

        for ldapName in self._schema.getManagedAttributeNames():

            schema = self._schema.getAttributeSchema( ldapName )
            if ldapName in readIds: 
		schema.addReaders( roleName )
            elif roleName in schema.getReaders():
                schema.delReaders( roleName )

            if ldapName in writeIds:
                schema.addWriters( roleName )
            elif roleName in schema.getWriters():
                schema.delWriters( roleName )

        return "Saved readers and writers settings for the %s role in each managed attribute" % roleName

    def _saveBrowse(self):

        # Save the browsing sort key and formatting string

	self._setBrowseFormat()
	self._setBrowseSortKey()
        return "Saved changes."

    def _setBrowseFormat(self):

	# Set the browsing format string

        browseFormat = self.request.get( 'browseFormat', '' )
        if browseFormat:
            self._schema._browseFormat = browseFormat

    def _setBrowseSortKey(self):
        browseSortKey = self.request.get( 'browseSortKey' )
        if browseSortKey in self._schema.getBrowseKeys():
            self._schema.setBrowseSortKey(browseSortKey)
        else:
            self._schema.setBrowseSortKey(self._schema.getBrowseKeys()[0])


    def _setNamingAttribute(self):

        # Set the naming attribute for this schema.

        namingAttribute = self.request.get( 'namingAttribute' )
        self._schema.setNamingAttribute( namingAttribute )

        return "Set naming attribute."

    def _searchFilterSave(self):

        # Save search filter specified in REQUEST

        searchFilter = self.request.get( 'searchFilter' )
        self._schema.setSearchFilter( searchFilter )

	return "Set search filter."

    def _searchKeysSave(self):

        # Save search keys specified in REQUEST

        searchSortKey = self.request.get( 'searchSortKey' )
        if searchSortKey in self._schema.getSearchKeys():
            self._schema.setSearchSortKey( searchSortKey )
        else:
            self._schema.setSearchSortKey(self._schema.getSearchKeys()[0])

        for key in self._schema.getSearchKeys():
            keyName = self.request.get( key+'_name', None )
            if keyName is not None:
                self._schema.setSearchKeyName( key, keyName )

        return "Saved changes."

    def _searchOrgUnitIndicatorSave(self):

        # Save browsing options specified in the request.

        if self.request.get( 'objectCategory' ) == 'person':
            personsOUAttr = self.request.get( 'personsOUAttr' )
            givenAttr = self.request.get( 'givenAttr' )
            if personsOUAttr == 'dn':
                self._schema.setPersonsOUAttr( 'dn' )
            elif personsOUAttr == 'useGivenAttr':
                self._schema.setPersonsOUAttr( givenAttr )

        return "Saved changes."

    def setPermissionName( self, permissionName ):

        # Set the name of the permission that controls the creation and
        # deletion of objects of this objectCategory in LDAP.

        self._schema.setPermissionName(permissionName)

    def zopeRoleNeedsANON(self, anonymousRoleName, roles):
	# If one of 'roles' is an acquired Zope role, then at the
	# directory level anonymous access must be allowed.  If 
	# anonymous access is not in 'roles,' then return the list of
	# acquired roles, so we can warn the user.  We add ANON when 
	# writing the security assertions in the ZReports class, not here.

	zope_roles = self.aq_parent.aq_parent.valid_roles()
	if anonymousRoleName in roles:
	    return []

	acquiredRoles = []
	for role in roles:
	    if role in zope_roles:
		acquiredRoles.append(role)

	return acquiredRoles

####################################################################
#
#   The method below is an auxiliary method that would be marked
#   as private class methods in other OO languages.  It does not
#   need any instance specific information and it should not be
#   "imported" into any other python module.
#
####################################################################

def getLDAPAttributeNames(objectClass, objectSchemas):
    """
        Returns lists of required and optional attributes for the
        given LDAP object class.
    """
    if not objectSchemas.has_key( lower(objectClass) ):
        return [],[]
    LDIFObject = objectSchemas[lower(objectClass)]
    required = LDIFObject.mustContain
    optional = LDIFObject.mayContain

    if LDIFObject.superClass:
        supRequired, supOptional = \
            getLDAPAttributeNames(LDIFObject.superClass, objectSchemas)
        required = mergeLists( required, supRequired )
        optional = mergeLists( optional, supOptional )

    return required, optional

def getLDAPSuperClassNames( objectClass, objectSchemas ):
    """
        Returns a list of superclass names for the given LDAP object
        class.
    """
    LDIFObject = objectSchemas[lower(objectClass)]

    if LDIFObject.superClass:
        super = [ LDIFObject.superClass, ]
        super.extend( getLDAPSuperClassNames( LDIFObject.superClass,
                                              objectSchemas ) )
        return super

    else:
        return []


