"""
    Testing the LDAPObject class.

    TestLDAPObject.py
    $Author: payne $
    $Date: 2003/12/07 04:13:53 $

Copyright (C) 2002 Association of Universities for Research in
Astronomy (AURA).  See LICENSE.txt for the usual disclaimers.

"""

__version__ = '$Revision: 1.34 $'[11:-1]

import sys
sys.path.insert(0,'..')

import unittest

from Connection import AuthenticatedUser
from LDAPObject import LDAPObject
from LDAPUtilities import *
from LDAPExceptions import *
from ManagedRoles import ManagedRoles
from ManagedSchema import ManagedSchema
from SchemaDefaults_test import *
from TestDirectory import getTestDirectory

class testLDAPObject(unittest.TestCase):

    """ Tests the Schema class in the Schema module """

    def setUp(self):
        """ Logon Test """

        directory = getTestDirectory()      
        roles = ManagedRoles()
        self.user = AuthenticatedUser('testadmin', 'testadmin',
                       directory)
        self.ldapObject = LDAPObject(self.user)

    def test01(self):
        """ 01 testLDAPObject: Retrieve an object """
        distinguishedName='uid=usera,ou=people,o=test'
        dn = self.ldapObject.refresh(distinguishedName)
        self.failUnless(DNsAreEquivalent( dn,distinguishedName ) )

    def test02(self):
        """ 02 testLDAPObject: testing getManagedAttributeNames """
        personalAttrs = ['carLicense',
                         'cn',
                         'givenName',
                         'homePostalAddress',
                         'homePhone',
                         'initials',
                         'sn']
        #personalAttrs = ['cn', 'givenName', 'initials', 'sn']
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        list = self.ldapObject.getManagedAttributeNames('Personal')
        self.failUnless( listsAreEquivalent( list, personalAttrs) )

    def test03(self):
        """ O3 testLDAPObject: testing updateDirectory modifying/adding an attribute """
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        cn = self.ldapObject.getManagedAttribute('cn')
        cn.setValues(['New Common Name'])
        self.ldapObject.updateDirectory()
        self.ldapObject.refresh()
        cn = self.ldapObject.getManagedAttribute('cn')
        self.failUnless( cn.getValues() == ['New Common Name'] )
        cn.setValues(['User, Anne'])
        self.ldapObject.updateDirectory()

    def test04(self):
        """ 04 testLDAPObject: testing updateDirectory - deleting an attribute: """
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        init = self.ldapObject.getManagedAttribute('initials')
        init.setValues(['A'])
        self.ldapObject.updateDirectory()
        
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        init = self.ldapObject.getManagedAttribute('initials')
        self.failUnless( init.getValues() == ['A'] )
        
        init.setValues([])
        self.ldapObject.updateDirectory()
        
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        init = self.ldapObject.getManagedAttribute('initials')
        self.failUnless( init.getValues() == [] )

    def test05(self):
        """ 05 testLDAPObject: testing canReadAttribute """
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        self.failUnless(self.ldapObject.canReadAttribute('cn'))

    def test06(self):
        """ 06 testLDAPObject: testing canWriteAttribute """
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        cn = self.ldapObject.getManagedAttribute('cn')
        self.failUnless(self.ldapObject.canWriteAttribute('cn'))

    def test07(self):
        """ 07 testLDAPObject: testing getRequiredObjectClasses """
        requiredValues = ['inetOrgPerson',
                          'organizationalPerson',
                          'person',
                          'top']
        self.ldapObject.refresh(distinguishedName='uid=usera,ou=people,o=test')
        self.failUnless( listsAreEquivalent( requiredValues,
                                             self.ldapObject.getRequiredObjectClasses()))

    def test08(self):
        """ 08 testLDAPObject: delete an object """
        dn = 'cn=New Group,ou=Groups,o=test'
        testGroup = LDAPObject( self.user)
        testGroup.create(distinguishedName=dn, objectCategory='group')
        testGroup.delete(TRUE)
        try:
            testGroup = LDAPObject( self.user)
            testGroup.refresh(distinguishedName=dn)
        except NO_SUCH_OBJECT:
            self.failUnless(TRUE)
        else:
            self.fail()

    def test09(self):
        """ 09 testLDAPObject: delete a non empty object """
        dn1 = 'ou=bogus test group,ou=Permanent Test OU,o=test'
        dn2 = 'ou=another group,' + dn1
        testOU1 = LDAPObject( self.user )
        testOU1.create(distinguishedName=dn1, objectCategory='orgunit')
        testOU2 = LDAPObject( self.user )
        testOU2.create(distinguishedName=dn2, objectCategory='orgunit')
        testOU1.delete(TRUE)
        try:
            testOU = LDAPObject( self.user )
            testOU.refresh(distinguishedName = dn1)
        except NO_SUCH_OBJECT:
            self.failUnless(TRUE)
        else:
            self.fail()

    def test10(self):
        """ 10 testLDAPObject: rename an object """
        
        oldDN = 'uid=usera,ou=people,o=test'
        newDN = 'uid=userX,ou=people,o=test'
    
        self.ldapObject.refresh( oldDN )
        dn = self.ldapObject.rename( 'userX' )
        self.failUnless( DNsAreEquivalent( dn, newDN ) )
        dn = self.ldapObject.rename( 'usera' )
        self.failUnless( DNsAreEquivalent( dn, oldDN ) )

    def test11(self):
        """ 11 testLDAPObject: rename an object with collision """
        
        oldDN = 'uid=usera,ou=people,o=test'
        self.ldapObject.refresh( oldDN )
        self.assertRaises( RENAME_ERROR, self.ldapObject.rename, 'userb' )

    def test12(self):
        """ 12 testLDAPObject: move an object to a new superior """

        oldDN = 'uid=usera,ou=people,o=test'
        self.ldapObject.refresh( oldDN )
        try:
            dn = self.ldapObject.rename( 'usera', 'o=test' )
        except NOT_SUPPORTED:
            return
        self.failUnless( DNsAreEquivalent( dn, 'uid=usera,o=test' ) )
        self.ldapObject.refresh( dn )
        dn = self.ldapObject.rename( 'usera', 'ou=people,o=test' )
        self.failUnless( DNsAreEquivalent( dn, oldDN ) )
        
    def tearDown(self):
        self.user.unbind()

if __name__ == '__main__':
    
    tests = unittest.makeSuite(testLDAPObject, 'test')
    unittest.TextTestRunner(verbosity=2).run(tests)
